import logging
import socket
from urllib.parse import urlparse

import requests
from task_processing.runners.subscription import Subscription
from task_processing.task_processor import TaskProcessor
from twisted.internet.defer import logError

import tron.metrics as metrics
from tron.actioncommand import ActionCommand
from tron.utils.queue import PyDeferredQueue

TASK_LOG_FORMAT = '%(asctime)s %(name)s %(levelname)s %(message)s'
TASK_OUTPUT_LOGGER = 'tron.mesos.task_output'

log = logging.getLogger(__name__)


def get_mesos_leader(master_address, mesos_master_port):
    url = "%s:%s/redirect" % (master_address, mesos_master_port)
    response = requests.get(url)
    return '{}:{}'.format(urlparse(response.url).hostname, mesos_master_port)


def combine_volumes(defaults, overrides):
    """Helper to reconcile lists of volume mounts.

    If any volumes have the same container path, the one in overrides wins.
    """
    result = {mount['container_path']: mount for mount in defaults}
    for mount in overrides:
        result[mount['container_path']] = mount
    return list(result.values())


def get_secret_from_file(file_path):
    if file_path is not None:
        with open(file_path) as f:
            secret = f.read().strip()
    else:
        secret = None
    return secret


class MesosClusterRepository:
    """A class that stores MesosCluster objects and configuration."""

    # Config values
    mesos_enabled = False
    master_address = None
    master_port = None
    secret_file = None
    role = None
    principal = None
    default_volumes = ()
    dockercfg_location = None
    offer_timeout = None
    secret = None

    name = 'frameworks'
    clusters = {}
    state_data = {}
    state_watcher = None

    @classmethod
    def attach(cls, _, observer):
        cls.state_watcher = observer

    @classmethod
    def get_cluster(cls, master_address=None):
        if master_address is None:
            master_address = cls.master_address
        if master_address not in cls.clusters:
            framework_id = cls.state_data.get(master_address)
            cluster = MesosCluster(
                mesos_address=master_address,
                mesos_master_port=cls.master_port,
                secret=cls.secret,
                principal=cls.principal,
                mesos_role=cls.role,
                framework_id=framework_id,
                enabled=cls.mesos_enabled,
                default_volumes=cls.default_volumes,
                dockercfg_location=cls.dockercfg_location,
                offer_timeout=cls.offer_timeout,
            )
            cls.clusters[master_address] = cluster
        return cls.clusters[master_address]

    @classmethod
    def shutdown(cls):
        for cluster in cls.clusters.values():
            cluster.stop()

    @classmethod
    def configure(cls, mesos_options):
        cls.master_address = mesos_options.master_address
        cls.master_port = mesos_options.master_port
        cls.secret_file = mesos_options.secret_file
        cls.role = mesos_options.role
        cls.secret = get_secret_from_file(cls.secret_file)
        cls.principal = mesos_options.principal
        cls.mesos_enabled = mesos_options.enabled
        cls.default_volumes = [
            vol._asdict() for vol in mesos_options.default_volumes
        ]
        cls.dockercfg_location = mesos_options.dockercfg_location
        cls.offer_timeout = mesos_options.offer_timeout

        for cluster in cls.clusters.values():
            cluster.set_enabled(cls.mesos_enabled)
            cluster.configure_tasks(
                default_volumes=cls.default_volumes,
                dockercfg_location=cls.dockercfg_location,
                offer_timeout=cls.offer_timeout,
            )

    @classmethod
    def restore_state(cls, mesos_state):
        cls.state_data = mesos_state.get(cls.name, {})

    @classmethod
    def save(cls, master_address, framework_id):
        cls.state_data[master_address] = framework_id
        cls.state_watcher.handler(cls, None)

    @classmethod
    def remove(cls, master_address):
        if master_address in cls.state_data:
            del cls.state_data[master_address]
            cls.state_watcher.handler(cls, None)


class MesosTask(ActionCommand):
    ERROR_STATES = frozenset(['failed', 'killed', 'error'])

    def __init__(self, id, task_config, serializer=None):
        super(MesosTask, self).__init__(id, task_config.cmd, serializer)
        self.task_config = task_config

        self.log = self.get_event_logger()
        self.setup_output_logging()

        self.log.info(
            'Mesos task {} created with config {}'.format(
                self.get_mesos_id(),
                self.get_config(),
            ),
        )

    def get_event_logger(self):
        log = logging.getLogger(__name__ + '.' + self.id)
        # Every time a task gets created, this function runs and will add
        # more stderr handlers to the logger, which results in duplicate log
        # output. We only want to add the stderr handler if the logger does not
        # have a handler yet.
        if not len(log.handlers):
            handler = logging.StreamHandler(self.stderr)
            handler.setFormatter(logging.Formatter(TASK_LOG_FORMAT))
            log.addHandler(handler)
        return log

    def setup_output_logging(self):
        task_id = self.get_mesos_id()
        stdout_logger = logging.getLogger(
            '{}.{}.{}'.format(TASK_OUTPUT_LOGGER, task_id, 'stdout'),
        )
        stdout_logger.addHandler(logging.StreamHandler(self.stdout))
        stderr_logger = logging.getLogger(
            '{}.{}.{}'.format(TASK_OUTPUT_LOGGER, task_id, 'stderr'),
        )
        stderr_logger.addHandler(logging.StreamHandler(self.stderr))

    def get_mesos_id(self):
        return self.task_config.task_id

    def get_config(self):
        return self.task_config

    def report_resources(self, decrement=False):
        multiplier = -1 if decrement else 1
        metrics.count(f'tron.mesos.cpus', self.task_config.cpus * multiplier)
        metrics.count(f'tron.mesos.mem', self.task_config.mem * multiplier)
        metrics.count(f'tron.mesos.disk', self.task_config.disk * multiplier)

    def log_event_info(self, event):
        # Separate out so task still transitions even if this nice-to-have logging fails.
        mesos_type = getattr(event, 'platform_type', None)
        if mesos_type == 'staging':
            # TODO: Save these in state?
            agent = event.raw.get('offer', {}).get('agent_id', {}).get('value')
            hostname = event.raw.get('offer', {}).get('hostname')
            self.log.info(
                f'Staging task on agent {agent} (hostname {hostname})'
            )
        elif mesos_type == 'running':
            agent = event.raw.get('agent_id', {}).get('value')
            self.log.info(f'Running on agent {agent}')
        elif mesos_type == 'finished':
            pass
        elif mesos_type in self.ERROR_STATES:
            self.log.error(f'Error from Mesos: {event.raw}')
        elif mesos_type is None:
            self.log.info(f'Non-Mesos event: {event.raw}')

        # Mesos events may have task reasons
        if mesos_type:
            message = event.raw.get('message', '')
            reason = event.raw.get('reason', '')
            if message or reason:
                self.log.info(f'More info: {reason}: {message}')

    def handle_event(self, event):
        event_id = getattr(event, 'task_id', None)
        if event_id != self.get_mesos_id():
            self.log.warning(
                'Event task id {} does not match, ignoring'.format(event_id),
            )
            return
        mesos_type = getattr(event, 'platform_type', None)

        self.log.info(
            'Got event for task {id}, Mesos type {type}'.format(
                id=event_id,
                type=mesos_type,
            )
        )
        try:
            self.log_event_info(event)
        except Exception as e:
            self.log.warning('Exception while logging event: {}'.format(e))

        if mesos_type == 'staging':
            pass
        elif mesos_type == 'running':
            self.started()
        elif mesos_type == 'finished':
            self.exited(0)
        elif mesos_type == 'lost':
            self.exited(None)
        elif mesos_type in self.ERROR_STATES:
            self.exited(1)
        elif mesos_type is None:
            pass
        else:
            self.log.info(
                'Did not handle unknown type of event: {}'.format(event),
            )

        if event.terminal:
            self.log.info('Event was terminal, closing task')
            self.report_resources(decrement=True)

            exit_code = int(not getattr(event, 'success', False))
            # Returns False if we've already exited normally above
            unexpected_error = self.exited(exit_code)
            if unexpected_error:
                self.log.error('Unexpected failure, exiting')

            self.done()


class MesosCluster:
    def __init__(
        self,
        mesos_address,
        mesos_master_port=None,
        secret=None,
        principal=None,
        mesos_role=None,
        framework_id=None,
        enabled=True,
        default_volumes=None,
        dockercfg_location=None,
        offer_timeout=None,
    ):
        self.mesos_address = mesos_address
        self.mesos_master_port = mesos_master_port
        self.secret = secret
        self.principal = principal
        self.mesos_role = mesos_role
        self.enabled = enabled
        self.default_volumes = default_volumes or []
        self.dockercfg_location = dockercfg_location
        self.offer_timeout = offer_timeout
        self.framework_id = framework_id

        self.processor = TaskProcessor()
        self.queue = PyDeferredQueue()
        self.deferred = None
        self.runner = None
        self.tasks = {}

        self.processor.load_plugin(
            provider_module='task_processing.plugins.mesos'
        )
        self.connect()

    def set_enabled(self, is_enabled):
        self.enabled = is_enabled
        if is_enabled:
            self.connect()
        else:
            self.stop(fail_tasks=True)

    def configure_tasks(
        self,
        default_volumes,
        dockercfg_location,
        offer_timeout,
    ):
        self.default_volumes = default_volumes
        self.dockercfg_location = dockercfg_location
        self.offer_timeout = offer_timeout

    def connect(self):
        self.runner = self.get_runner(self.mesos_address, self.queue)
        self.handle_next_event()

    def handle_next_event(self, deferred_result=None):
        if self.deferred and not self.deferred.called:
            log.warning(
                'Already have handlers waiting for next event in queue, '
                'not adding more'
            )
            return
        self.deferred = self.queue.get()
        self.deferred.addCallback(self._process_event)
        self.deferred.addCallback(self.handle_next_event)
        self.deferred.addErrback(logError)
        self.deferred.addErrback(self.handle_next_event)

    def _check_connection(self):
        if self.runner.stopping:
            # Last framework was terminated for some reason, re-connect.
            log.info('Last framework stopped, re-connecting')
            self.connect()
        elif self.deferred.called:
            # Just in case callbacks are missing, re-add.
            self.handle_next_event()

    def submit(self, task):
        if not task:
            return

        if not self.enabled:
            task.log.info('Task failed to start, Mesos is disabled.')
            task.exited(1)
            return
        self._check_connection()

        mesos_task_id = task.get_mesos_id()
        self.tasks[mesos_task_id] = task
        self.runner.run(task.get_config())
        log.info(
            'Submitting task {} to {}'.format(
                mesos_task_id,
                self.mesos_address,
            ),
        )
        task.report_resources()

    def recover(self, task):
        if not task:
            return

        if not self.enabled:
            task.log.info('Could not recover task, Mesos is disabled.')
            task.exited(None)
            return
        self._check_connection()

        mesos_task_id = task.get_mesos_id()
        self.tasks[mesos_task_id] = task
        task.log.info('Reconciling state for this task from Mesos')
        task.started()
        self.runner.reconcile(task.get_config())
        task.report_resources()

    def create_task(
        self,
        action_run_id,
        command,
        cpus,
        mem,
        disk,
        constraints,
        docker_image,
        docker_parameters,
        env,
        extra_volumes,
        serializer,
        task_id=None,
    ):
        if not self.runner:
            return None

        uris = [self.dockercfg_location] if self.dockercfg_location else []
        volumes = combine_volumes(self.default_volumes, extra_volumes)
        task_kwargs = {
            'name': action_run_id,
            'cmd': command,
            'cpus': cpus,
            'mem': mem,
            'disk': disk,
            'constraints': constraints,
            'image': docker_image,
            'docker_parameters': docker_parameters,
            'environment': env,
            'volumes': volumes,
            'uris': uris,
            'offer_timeout': self.offer_timeout,
        }
        task_config = self.runner.TASK_CONFIG_INTERFACE(**task_kwargs)

        if task_id is not None:
            try:
                task_config = task_config.set_task_id(task_id)
            except ValueError:
                log.error(f'Invalid {task_id} for {action_run_id}')
                return

        return MesosTask(action_run_id, task_config, serializer)

    def get_runner(self, mesos_address, queue):
        if not self.enabled:
            log.info('Mesos is disabled, not creating a framework.')
            return None

        if self.runner and not self.runner.stopping:
            log.info('Already have a running framework, not creating one.')
            return self.runner

        framework_name = 'tron-{}'.format(socket.gethostname())
        executor = self.processor.executor_from_config(
            provider='mesos_task',
            provider_config={
                'secret':
                    self.secret,
                'principal':
                    self.principal,
                'mesos_address':
                    get_mesos_leader(mesos_address, self.mesos_master_port),
                'role':
                    self.mesos_role,
                'framework_name':
                    framework_name,
                'framework_id':
                    self.framework_id,
                'failover':
                    True,
            }
        )

        def log_output(task_id, message, stream):
            logger = logging.getLogger(
                '{}.{}.{}'.format(
                    TASK_OUTPUT_LOGGER,
                    task_id,
                    stream,
                )
            )
            logger.info(message)

        logging_executor = self.processor.executor_from_config(
            provider='logging',
            provider_config={
                'downstream_executor': executor,
                'handler': log_output,
                'format_string': '{line}',
            },
        )
        return Subscription(logging_executor, queue)

    def _process_event(self, event):
        if event.kind == 'control':
            message = getattr(event, 'message', None)
            if message == 'stop':
                # Framework has been removed, stop it.
                log.warning('Framework has been stopped: {}'.format(event.raw))
                self.stop()
                MesosClusterRepository.remove(self.mesos_address)
            elif message == 'unknown':
                log.warning(
                    'Unknown error from Mesos master: {}'.format(event.raw)
                )
            elif message == 'registered':
                framework_id = event.raw['framework_id']['value']
                MesosClusterRepository.save(self.mesos_address, framework_id)
            else:
                log.warning('Unknown type of control event: {}'.format(event))

        elif event.kind == 'task':
            if not hasattr(event, 'task_id'):
                log.warning('Task event missing task_id: {}'.format(event))
                return
            if event.task_id not in self.tasks:
                log.warning(
                    'Received event for unknown task {}: {}'.format(
                        event.task_id,
                        event,
                    ),
                )
                return
            task = self.tasks[event.task_id]
            task.handle_event(event)
            if task.is_done:
                del self.tasks[event.task_id]
        else:
            log.warning('Unknown type of event: {}'.format(event))

    def stop(self, fail_tasks=False):
        self.framework_id = None
        if self.runner:
            self.runner.stop()

        # Clear message queue
        if self.deferred:
            self.deferred.cancel()
            self.deferred = None
        self.queue = PyDeferredQueue()

        if fail_tasks:
            for key, task in list(self.tasks.items()):
                task.exited(None)
                del self.tasks[key]

    def kill(self, task_id):
        return self.runner.kill(task_id)



from __future__ import absolute_import
from __future__ import unicode_literals

import contextlib
import fcntl
import logging
import os
import signal

log = logging.getLogger(__name__)


def maybe_decode(maybe_string):
    if type(maybe_string) is bytes:
        return maybe_string.decode()
    return maybe_string


def maybe_encode(maybe_bytes):
    if type(maybe_bytes) is not bytes:
        return maybe_bytes.encode()
    return maybe_bytes


def next_or_none(iterable):
    try:
        return next(iterable)
    except StopIteration:
        pass


@contextlib.contextmanager
def flock(fd):
    close = False
    if isinstance(fd, str):
        fd = open(fd, 'a')
        close = True

    try:
        fcntl.lockf(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except BlockingIOError as e:  # locked by someone else
        log.debug(f"Locked by another process: {fd}")
        raise e

    try:
        yield
    finally:
        fcntl.lockf(fd, fcntl.LOCK_UN)
        if close:
            fd.close()


@contextlib.contextmanager
def chdir(path):
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(cwd)


@contextlib.contextmanager
def signals(signal_map):
    orig_map = {}
    for signum, handler in signal_map.items():
        orig_map[signum] = signal.signal(signum, handler)

    try:
        yield
    finally:
        for signum, handler in orig_map.items():
            signal.signal(signum, handler)



#!/usr/bin/env python3.6
import datetime
import logging
import pprint
import sys
import time
from collections import defaultdict
from enum import Enum

import pytimeparse
from pyrsistent import m
from pyrsistent import pmap
from pysensu_yelp import send_event

from tron.commands import cmd_utils
from tron.commands import display
from tron.commands.client import Client
from tron.commands.client import get_object_type_from_identifier

PRECIOUS_JOB_ATTR = 'check_that_every_day_has_a_successful_run'

log = logging.getLogger('check_tron_jobs')

_run_interval = None


class State(Enum):
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    STUCK = "stuck"
    NO_RUN_YET = "no_run_yet"
    NOT_SCHEDULED = "not_scheduled"
    NO_RUNS_TO_CHECK = "no_runs_to_check"
    UNKNOWN = "unknown"
    SKIPPED = "skipped"


def parse_cli():
    parser = cmd_utils.build_option_parser()
    parser.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Don't actually send alerts out. Defaults to %(default)s",
    )
    parser.add_argument(
        "--job",
        default=None,
        help="Check a particular job. If unset checks all jobs",
    )
    parser.add_argument(
        "--run-interval",
        help="Run interval for this monitoring script. This is used to "
        "calculate realert and alert_after setting. "
        "Default to %(default)s (seconds)",
        type=int,
        dest="run_interval",
        default=300,
    )
    args = parser.parse_args()
    return args


def _timestamp_to_timeobj(timestamp):
    return time.strptime(timestamp, '%Y-%m-%d %H:%M:%S')


def _timestamp_to_shortdate(timestamp, separator='.'):
    return time.strftime(
        '%Y{0}%m{0}%d'.format(separator),
        _timestamp_to_timeobj(timestamp),
    )


def compute_check_result_for_job_runs(client, job, job_content):
    url_index = client.index()
    kwargs = {}
    if job_content is None:
        kwargs["output"] = "OK: {} was just added and hasn't run yet.".format(
            job['name'],
        )
        kwargs["status"] = 0
        return kwargs

    relevant_job_run, last_state = get_relevant_run_and_state(job_content)
    if relevant_job_run is None:
        kwargs["output"] = f"CRIT: {job['name']} hasn't had a successful " \
            f"run yet.\n{pretty_print_job(job_content)}"
        kwargs["status"] = 2
        return kwargs
    else:  # if no run scheduled, no run_time available
        relevant_job_run_date = _timestamp_to_shortdate(
            relevant_job_run['run_time']
        )

    # A job_run is like MASTER.foo.1
    job_run_id = get_object_type_from_identifier(
        url_index,
        relevant_job_run['id'],
    )
    action_runs = client.job(job_run_id.url, include_action_runs=True)
    # A job action is like MASTER.foo.1.step1
    actions_expected_runtime = job_content.get('actions_expected_runtime', {})
    relevant_action = get_relevant_action(
        action_runs=action_runs["runs"],
        last_state=last_state,
        actions_expected_runtime=actions_expected_runtime
    )
    action_run_id = get_object_type_from_identifier(
        url_index,
        relevant_action['id'],
    )
    action_run_details = client.action_runs(action_run_id.url, num_lines=10)

    if last_state == State.SUCCEEDED:
        prefix = "OK: The last job run succeeded"
        status = 0
    elif last_state == State.NO_RUNS_TO_CHECK:
        prefix = "OK: The job is 'new' and/or has no runs to check"
        status = 0
    elif last_state == State.SKIPPED:
        prefix = "OK: The last job run was skipped"
        status = 0
    elif last_state == State.STUCK:
        prefix = "WARN: Job exceeded expected runtime or still running when next job is scheduled"
        status = 1
    elif last_state == State.FAILED:
        prefix = "CRIT: The last job run failed!"
        status = 2
    elif last_state == State.NOT_SCHEDULED:
        prefix = "CRIT: Job is not scheduled at all!"
        status = 2
    elif last_state == State.UNKNOWN:
        prefix = "CRIT: Job has gone 'unknown' and might need manual intervention"
        status = 2
    else:
        prefix = "UNKNOWN: The job is in a state that check_tron_jobs doesn't understand"
        status = 3

    precious_runs_note = ''
    if job['monitoring'].get(PRECIOUS_JOB_ATTR, False) and status != 0:
        precious_runs_note = f"Note: This alert is the run for {relevant_job_run_date}. A resolve event will not occur until a job run for this date succeeds.\n"

    kwargs["output"] = (
        f"{prefix}\n"
        f"{job['name']}'s latest run for {relevant_job_run_date} ({relevant_job_run['id']}) {relevant_job_run['state']}\n"
        f"{precious_runs_note}"
        "\nHere is the last action:\n"
        f"{pretty_print_actions(action_run_details)}\n\n"
        "And the job run view:\n"
        f"{pretty_print_job_run(relevant_job_run)}\n\n"
        "Here is the whole job view for context:\n"
        f"{pretty_print_job(job_content)}"
    )
    kwargs["status"] = status
    return kwargs


def pretty_print_job(job_content):
    return display.format_job_details(job_content)


def pretty_print_job_run(job_run):
    display_action = display.DisplayActionRuns()
    return display_action.format(job_run)


def pretty_print_actions(action_run):
    return display.format_action_run_details(action_run)


def get_relevant_run_and_state(job_content):
    # The order of job run to check is as follows:
    #   1. The scheduled but hasn't run one checked first
    #   2. Then currently running ones are always checked (in case an action is failed/unknown)
    #   3. If there are multiple running ones, then most recent run_time wins
    #   4. If nothing is currently running, then most recent end_time wins
    job_runs = sorted(
        job_content.get('runs', []),
        key=lambda k: (k['end_time'] is None, k['end_time'], k['run_time']),
        reverse=True,
    )
    if len(job_runs) == 0:
        return None, State.NO_RUN_YET
    run = is_job_scheduled(job_runs)
    # If runs are precious, then it is possible for a day to have no scheduled
    # run if it already had a successful one. Thus, we do not want to return a
    # NOT_SCHEDULED state, but a SUCCEEDED.
    if run is None and not job_content['monitoring'
                                       ].get(PRECIOUS_JOB_ATTR, False):
        return job_runs[0], State.NOT_SCHEDULED
    job_expected_runtime = job_content.get('expected_runtime', None)
    actions_expected_runtime = job_content.get('actions_expected_runtime', {})
    run = is_job_stuck(
        job_runs=job_runs,
        job_expected_runtime=job_expected_runtime,
        actions_expected_runtime=actions_expected_runtime,
        allow_overlap=job_content.get('allow_overlap'),
    )
    if run is not None:
        return run, State.STUCK
    for run in job_runs:
        state = run.get('state', 'unknown')
        if state in ["failed", "succeeded", "unknown", "skipped"]:
            return run, State(state)
        elif state in ["running"]:
            action_state = is_action_failed_or_unknown(run)
            if action_state != State.SUCCEEDED:
                return run, action_state
    return job_runs[0], State.NO_RUNS_TO_CHECK


def is_action_failed_or_unknown(job_run):
    for run in job_run.get('runs', []):
        if run.get('state', None) in ["failed", "unknown"]:
            return State(run.get('state'))
    return State.SUCCEEDED


def is_job_scheduled(job_runs):
    for job_run in job_runs:
        if job_run.get('state', 'unknown') in ["scheduled", "queued"]:
            return job_run
    return None


def is_job_stuck(
    job_runs, job_expected_runtime, actions_expected_runtime, allow_overlap
):
    next_run_time = None
    for job_run in job_runs:
        if job_run.get('state', 'unknown') == "running":
            if is_job_run_exceeding_expected_runtime(
                job_run, job_expected_runtime
            ):
                return job_run
            # check if it is still running at next scheduled job run time
            if not allow_overlap and next_run_time:
                difftime = _timestamp_to_timeobj(next_run_time)
                if time.time() > time.mktime(difftime):
                    return job_run
            for action_run in job_run.get('runs', []):
                if is_action_run_exceeding_expected_runtime(
                    action_run, actions_expected_runtime
                ):
                    return job_run

        next_run_time = job_run.get('run_time', None)
    return None


def is_job_run_exceeding_expected_runtime(job_run, job_expected_runtime):
    if job_expected_runtime is not None and job_run.get(
        'state', 'unknown'
    ) == "running":
        duration_seconds = pytimeparse.parse(job_run.get('duration', ''))
        if duration_seconds > job_expected_runtime:
            return True
    return False


def is_action_run_exceeding_expected_runtime(
    action_run, actions_expected_runtime
):
    if action_run.get('state', 'unknown') == 'running':
        action_name = action_run.get('action_name', None)
        if action_name in actions_expected_runtime and actions_expected_runtime[
            action_name
        ] is not None:
            duration_seconds = pytimeparse.parse(
                action_run.get('duration', '')
            )
            if duration_seconds > actions_expected_runtime[action_name]:
                return True
    return False


def get_relevant_action(*, action_runs, last_state, actions_expected_runtime):
    stuck_action_run_candidate = None
    for action_run in reversed(action_runs):
        action_state = action_run.get('state', 'unknown')
        try:
            if State(action_state) == last_state:
                return action_run
        except ValueError:
            if last_state == State.STUCK:
                if is_action_run_exceeding_expected_runtime(
                    action_run, actions_expected_runtime
                ):
                    return action_run
                if action_state == 'running':
                    stuck_action_run_candidate = action_run
    return stuck_action_run_candidate or action_runs[-1]


def guess_realert_every(job):
    try:
        job_next_run = job.get('next_run', None)
        if job_next_run is None:
            return -1
        job_runs = job.get('runs', [])
        job_runs_started = [
            run for run in job_runs if run['start_time'] is not None
        ]
        if len(job_runs_started) == 0:
            return -1
        job_previous_run = max(
            job_runs_started,
            key=lambda k: k['start_time'],
        ).get('start_time')
        time_diff = (
            time.mktime(_timestamp_to_timeobj(job_next_run)) -
            time.mktime(_timestamp_to_timeobj(job_previous_run))
        )
        realert_every = max(int(time_diff / _run_interval), 1)
    except Exception as e:
        log.warning("guess_realert_every failed: {}".format(e))
        return -1
    return realert_every


def sort_runs_by_interval(job_content, interval='day', until=None):
    """ Sorts a job's runs by a time interval (day, hour, minute, or second),
    according to a job run's run time.
    """
    interval_formats = {
        'day': '%Y.%m.%d',
        'hour': '%Y.%m.%d-%H',
        'minute': '%Y.%m.%d-%H.%M',
        'second': '%Y.%m.%d-%H.%M.%S',
    }

    run_buckets = defaultdict(list)
    if job_content is not None:
        if not until:
            until = time.time()  # can't set in default arg
        if job_content['runs']:
            earliest_run_time = min([
                time.mktime(_timestamp_to_timeobj(run['run_time']))
                for run in job_content['runs']
            ])
        else:
            earliest_run_time = until

        # We add all dates by interval between our earliest run_time and now,
        # allowing functions downstream to see if some dates had no runs
        start = datetime.datetime.fromtimestamp(earliest_run_time)
        end = datetime.datetime.fromtimestamp(until)
        step = datetime.timedelta(**{f'{interval}s': 1})
        while start <= end:
            run_buckets[start.strftime(interval_formats[interval])] = []
            start += step

        # Bucket runs by interval
        for run in job_content['runs']:
            # If interval is invalid, will raise a KeyError
            run_time = time.strftime(
                interval_formats[interval],
                _timestamp_to_timeobj(run['run_time']),
            )
            run_buckets[run_time].append(run)
    return dict(run_buckets)


def compute_check_result_for_job(client, job):
    kwargs = m(
        name="check_tron_job.{}".format(job['name']),
        source="tron",
    )
    if 'realert_every' not in kwargs:
        kwargs = kwargs.set('realert_every', guess_realert_every(job))
    kwargs = kwargs.set('check_every', f"{_run_interval}s")

    # We want to prevent a monitoring config from setting the check_every
    # attribute, since one config should not dictate how often this script runs
    sensu_kwargs = (
        pmap(job['monitoring']).discard(PRECIOUS_JOB_ATTR)
        .discard('check_every')
    )
    kwargs = kwargs.update(sensu_kwargs)

    kwargs_list = []
    if job["status"] == "disabled":
        kwargs = kwargs.set(
            'output',
            "OK: {} is disabled and won't be checked.".format(job['name'], )
        )
        kwargs = kwargs.set('status', 0)
        kwargs_list.append(kwargs)
    else:
        # The job is not disabled, therefore we have to look at its run history
        url_index = client.index()
        tron_id = get_object_type_from_identifier(url_index, job["name"])
        job_content = pmap(
            client.job(
                tron_id.url,
                count=20,
                include_action_runs=True,
            )
        )

        if job['monitoring'].get(PRECIOUS_JOB_ATTR, False):
            dated_runs = sort_runs_by_interval(job_content, interval='day')
        else:
            dated_runs = {'': job_content['runs']}

        for date, runs in dated_runs.items():
            results = compute_check_result_for_job_runs(
                job=job,
                job_content=job_content.set('runs', runs),
                client=client,
            )
            dated_kwargs = kwargs.update(results)
            if date:  # if empty date, leave job name alone
                dated_kwargs = dated_kwargs.set(
                    'name', f"{kwargs['name']}-{date}"
                )
            kwargs_list.append(dated_kwargs)

    return [dict(kws) for kws in kwargs_list]


def check_job(job, client):
    if job.get('monitoring', {}) == {}:
        log.debug(
            "Not checking {}, no monitoring metadata setup.".format(
                job['name'],
            )
        )
        return
    if job.get('monitoring').get('team', None) is None:
        log.debug("Not checking {}, no team specified".format(job['name']))
        return
    log.info("Checking {}".format(job['name']))
    return compute_check_result_for_job(job=job, client=client)


def check_job_result(job, client, dry_run):
    results = check_job(job, client)
    if not results:
        return

    for result in results:
        if dry_run:
            log.info("Would have sent this event to sensu: ")
            log.info(pprint.pformat(result))
        else:
            log.debug("Sending event: {}".format(pprint.pformat(result)))
            if 'runbook' not in result:
                result[
                    'runbook'
                ] = "No runbook specified. Please specify a runbook in the monitoring section of the job definition.",
            send_event(**result)


def main():
    args = parse_cli()
    cmd_utils.setup_logging(args)
    cmd_utils.load_config(args)
    client = Client(args.server)

    error_code = 0
    global _run_interval
    _run_interval = args.run_interval
    if args.job is None:
        jobs = client.jobs(include_job_runs=True)
        for job in jobs:
            try:
                check_job_result(job=job, client=client, dry_run=args.dry_run)
            except Exception as e:
                log.warning(
                    "check job result fails for job {}: {}".format(
                        job.get('name', ''),
                        e,
                    )
                )
                error_code = 1
    else:
        job_url = client.get_url(args.job)
        job = client.job_runs(job_url)
        check_job_result(job=job, client=client, dry_run=args.dry_run)

    return error_code


if __name__ == '__main__':
    sys.exit(main())



import logging
import os
import pickle
import signal
import time
from collections import defaultdict
from collections import deque

from twisted.internet import reactor

log = logging.getLogger(__name__)


def consume_dequeue(queue, func):
    queue_length = len(queue)
    for _ in range(queue_length):
        func(queue.popleft())


class EventBus:
    instance = None

    @staticmethod
    def create(log_dir):
        """Create log directory and link to current log if those don't
        already exist"""
        EventBus.shutdown()
        eb = EventBus(log_dir)

        if not os.path.exists(eb.log_dir):
            log.warning(f"creating {eb.log_dir}")
            os.mkdir(eb.log_dir)

        if not os.path.exists(eb.log_current) or not os.path.exists(
            os.readlink(eb.log_current)
        ):
            log.warning(f"creating {eb.log_current}")
            eb.sync_save_log("initial save")

        EventBus.instance = eb
        return eb

    @staticmethod
    def start():
        if not EventBus.instance:
            return
        return EventBus.instance._start()

    @staticmethod
    def shutdown():
        if not EventBus.instance:
            return
        EventBus.instance._shutdown()
        EventBus.instance = None

    @staticmethod
    def publish(event):
        if not EventBus.instance:
            return
        return EventBus.instance._publish(event)

    @staticmethod
    def subscribe(prefix, subscriber, callback):
        if not EventBus.instance:
            return
        return EventBus.instance._subscribe(prefix, subscriber, callback)

    @staticmethod
    def clear_subscriptions(subscriber):
        if not EventBus.instance:
            return
        return EventBus.instance._clear_subscriptions(subscriber)

    @staticmethod
    def has_event(event):
        if not EventBus.instance:
            return
        return EventBus.instance._has_event(event)

    @staticmethod
    def discard(event):
        if not EventBus.instance:
            return
        return EventBus.instance._discard(event)

    def __init__(self, log_dir):
        self.enabled = False
        self.event_log = {}
        self.event_subscribers = defaultdict(list)
        self.publish_queue = deque()
        self.subscribe_queue = deque()
        self.clear_subscription_queue = deque()
        self.log_dir = log_dir
        self.log_current = os.path.join(self.log_dir, "current")
        self.log_updates = 0
        self.log_last_save = 0
        self.log_save_interval = 60  # save every minute
        self.log_save_updates = 100  # save every 100 updates

    def _start(self):
        self.enabled = True
        log.info("starting")
        self.sync_load_log()
        reactor.callLater(0, self.sync_loop)

    def _shutdown(self):
        if self.enabled:
            self.enabled = False
            self.sync_save_log("shutdown")
            log.info("shutdown completed")

    def _publish(self, event):
        if isinstance(event, str):
            event = {'id': event}
        if isinstance(event, dict):
            self.publish_queue.append(event)
            log.debug(f"publish of {event['id']} enqueued")
            return True
        else:
            log.error(f"can't publish {event!r}, must be dict")
            return False

    def _discard(self, event) -> bool:
        if not self._has_event(event):
            return False
        del self.event_log[event]
        return True

    def _subscribe(self, prefix, subscriber, callback):
        self.subscribe_queue.append((prefix, subscriber, callback))
        log.debug(f"subscription ({prefix}, {subscriber}) enqueued")

    def _clear_subscriptions(self, subscriber):
        self.clear_subscription_queue.append(subscriber)
        log.debug(f"clearing subscriptions for {subscriber}")

    def _has_event(self, event_id):
        return event_id in self.event_log

    def sync_load_log(self):
        started = time.time()
        with open(self.log_current, 'rb') as f:
            self.event_log = pickle.load(f)
        duration = time.time() - started
        log.info(f"log read from disk, took {duration:.4}s")

    def sync_save_log(self, reason):
        started = time.time()
        new_file = os.path.join(self.log_dir, f"{int(started)}.pickle")
        try:
            with open(new_file, 'xb') as f:
                pickle.dump(self.event_log, f)
        except FileExistsError:
            log.exception(
                f"unable to dump the log, file {new_file} already exists, "
                f"too many updates/sec? current: {self.log_updates}, "
                f"threshold: {self.log_save_updates}"
            )
            return False

        # atomically replace `current` symlink
        tmplink = os.path.join(self.log_dir, "tmp")
        try:
            os.remove(tmplink)
        except FileNotFoundError:
            pass
        os.symlink(new_file, tmplink)
        os.replace(tmplink, self.log_current)

        duration = time.time() - started
        log.info(f"log dumped to disk because {reason}, took {duration:.4}s")
        return True

    def sync_loop(self):
        if not self.enabled:
            return

        try:
            self.sync_process()
        except Exception:
            log.error("eventbus exception:", exc_info=1)
            os.kill(os.getpid(), signal.SIGTERM)

        reactor.callLater(1, self.sync_loop)

    def sync_process(self):
        save_reason = None
        if time.time() > self.log_last_save + self.log_save_interval:
            if self.log_updates > 0:
                save_reason = (
                    f"{self.log_save_interval}s passed, "
                    f"{self.log_updates} updates"
                )
            else:
                self.log_last_save = time.time()
                log.debug("skipping save, no updates")
        elif self.log_updates > self.log_save_updates:
            save_reason = f"{self.log_save_updates} updates"

        if save_reason and self.sync_save_log(save_reason):
            self.log_last_save = time.time()
            self.log_updates = 0

        consume_dequeue(self.subscribe_queue, self.sync_subscribe)
        consume_dequeue(
            self.clear_subscription_queue, self.sync_clear_subscriptions
        )
        consume_dequeue(self.publish_queue, self.sync_publish)

    def sync_publish(self, event):
        event = pickle.loads(pickle.dumps(event))
        event_id = event['id']
        del event['id']
        if event_id in self.event_log:
            if self.event_log[event_id] != event:
                log.info(f"replacing event: {event_id}")
            else:
                log.debug(f"duplicate event: {event}")
                return

        self.event_log[event_id] = event
        self.log_updates += 1
        log.debug(f"event stored: {event_id} {event}")

        reactor.callLater(0, self.sync_notify, event_id)

    def sync_subscribe(self, prefix_subscriber_cb):
        prefix, subscriber, cb = prefix_subscriber_cb
        self.event_subscribers[prefix].append((subscriber, cb))
        log.debug(f"subscriber registered: {prefix_subscriber_cb}")

    def sync_unsubscribe(self, prefix_sub):
        prefix, sub = prefix_sub

        if prefix not in self.event_subscribers:
            log.debug(f"can't unsubscribe, not found for prefix {prefix}")
            return

        new_subs = [
            sub_cb for sub_cb in self.event_subscribers[prefix]
            if sub_cb[0] != sub
        ]
        if new_subs:
            self.event_subscribers[prefix] = new_subs
        else:
            del self.event_subscribers[prefix]
        log.debug(f"subscription removed: {prefix} / {sub}")

    def sync_clear_subscriptions(self, subscriber):
        new_subscriptions = defaultdict(list)
        removed = 0
        for prefix, subs in self.event_subscribers.items():
            for (sub, cb) in subs:
                if sub == subscriber:
                    removed += 1
                    continue
                new_subscriptions[prefix].append((sub, cb))
        self.event_subscribers = new_subscriptions

        if removed > 0:
            log.debug(f"subscriptions of {subscriber} removed: {removed}")

    def sync_notify(self, event_id):
        event = self.event_log[event_id]
        log.debug(f"notifying subscribers about {event_id}")
        for prefix, subscribers in self.event_subscribers.items():
            log.debug(f"check {prefix}: {event_id.startswith(prefix)}")
            if event_id.startswith(prefix):
                for (sub, cb) in subscribers:
                    log.debug(f"notifying {sub} about {event_id}")
                    reactor.callLater(0, cb, dict(id=event_id, **event))



import mock
import pytest

import tron.metrics as metrics


@pytest.fixture(autouse=True)
def all_metrics():
    with mock.patch.object(metrics, 'all_metrics', new=dict()) as mock_all:
        yield mock_all


def test_get_metric(all_metrics):
    timer = metrics.get_metric(
        'timer', 'api.requests', {'method': 'GET'}, mock.Mock()
    )
    same_timer = metrics.get_metric(
        'timer', 'api.requests', {'method': 'GET'}, mock.Mock()
    )
    other_timer = metrics.get_metric(
        'timer', 'api.requests', {'method': 'POST'}, mock.Mock()
    )
    metrics.get_metric('something', 'name', None, mock.Mock())
    assert timer == same_timer
    assert other_timer != timer
    assert len(all_metrics) == 3


@mock.patch('tron.metrics.get_metric', autospec=True)
def test_timer(mock_get_metric):
    test_metric = metrics.Timer()
    mock_get_metric.return_value = test_metric
    metrics.timer('my_metric', 110)
    metrics.timer('my_metric', 84)
    mock_get_metric.assert_called_with(
        'timer',
        'my_metric',
        None,
        mock.ANY,
    )
    result = metrics.view_timer(test_metric)
    assert result['count'] == 2


@mock.patch('tron.metrics.get_metric', autospec=True)
def test_count(mock_get_metric):
    test_metric = metrics.Counter()
    mock_get_metric.return_value = test_metric
    metrics.count('my_metric', 13)
    metrics.count('my_metric', -1)
    mock_get_metric.assert_called_with(
        'counter',
        'my_metric',
        None,
        mock.ANY,
    )
    result = metrics.view_counter(test_metric)
    assert result['count'] == 12


@mock.patch('tron.metrics.get_metric', autospec=True)
def test_meter(mock_get_metric):
    test_metric = metrics.Meter()
    mock_get_metric.return_value = test_metric
    metrics.meter('my_metric')
    metrics.meter('my_metric')
    mock_get_metric.assert_called_with(
        'meter',
        'my_metric',
        None,
        mock.ANY,
    )
    result = metrics.view_meter(test_metric)
    assert result['count'] == 2


@mock.patch('tron.metrics.get_metric', autospec=True)
def test_gauge(mock_get_metric):
    test_metric = metrics.SimpleGauge()
    mock_get_metric.return_value = test_metric
    metrics.gauge('my_metric', 23)
    metrics.gauge('my_metric', 102)
    mock_get_metric.assert_called_with(
        'gauge',
        'my_metric',
        None,
        mock.ANY,
    )
    result = metrics.view_gauge(test_metric)
    assert result['value'] == 102


@mock.patch('tron.metrics.get_metric', autospec=True)
def test_histogram(mock_get_metric):
    test_metric = metrics.Histogram()
    mock_get_metric.return_value = test_metric
    metrics.histogram('my_metric', 2)
    metrics.histogram('my_metric', 92)
    mock_get_metric.assert_called_with(
        'histogram',
        'my_metric',
        None,
        mock.ANY,
    )
    result = metrics.view_histogram(test_metric)
    assert result['count'] == 2


def test_view_all_metrics_empty():
    result = metrics.view_all_metrics()
    assert result == {
        'counter': [],
        'gauge': [],
        'histogram': [],
        'meter': [],
        'timer': [],
    }


def test_view_all_metrics():
    metrics.timer('a', 1)
    metrics.count('b', 9, dimensions={'method': 'GET'})
    metrics.meter('c')
    metrics.gauge('d', 3)
    metrics.histogram('e', 2)
    metrics.histogram('f', 3)
    result = metrics.view_all_metrics()

    assert len(result['timer']) == 1
    assert result['timer'][0]['name'] == 'a'

    assert len(result['counter']) == 1
    assert result['counter'][0]['name'] == 'b'
    assert result['counter'][0]['dimensions'] == {'method': 'GET'}

    assert len(result['meter']) == 1
    assert result['meter'][0]['name'] == 'c'

    assert len(result['gauge']) == 1
    assert result['gauge'][0]['name'] == 'd'

    assert len(result['histogram']) == 2
    names = set(metric['name'] for metric in result['histogram'])
    assert names == set(['e', 'f'])



"""
 tron.core.actionrun
"""
import datetime
import logging
from typing import List

from twisted.internet import reactor

from tron import command_context
from tron import node
from tron.actioncommand import ActionCommand
from tron.actioncommand import NoActionRunnerFactory
from tron.actioncommand import SubprocessActionRunnerFactory
from tron.config.config_utils import StringFormatter
from tron.config.schema import ExecutorTypes
from tron.core import action
from tron.eventbus import EventBus
from tron.mesos import MesosClusterRepository
from tron.serialize import filehandler
from tron.utils import maybe_decode
from tron.utils import proxy
from tron.utils import timeutils
from tron.utils.observer import Observable
from tron.utils.observer import Observer
from tron.utils.state import Machine

log = logging.getLogger(__name__)


class ActionRunFactory(object):
    """Construct ActionRuns and ActionRunCollections for a JobRun and
    ActionGraph.
    """

    @classmethod
    def build_action_run_collection(cls, job_run, action_runner):
        """Create an ActionRunGraph from an ActionGraph and JobRun."""
        action_run_map = {
            maybe_decode(name): cls.build_run_for_action(
                job_run,
                action_inst,
                action_runner,
            )
            for name, action_inst in job_run.action_graph.action_map.items()
        }
        return ActionRunCollection(job_run.action_graph, action_run_map)

    @classmethod
    def action_run_collection_from_state(
        cls,
        job_run,
        runs_state_data,
        cleanup_action_state_data,
    ):
        action_runs = [
            cls.action_run_from_state(job_run, state_data)
            for state_data in runs_state_data
        ]
        if cleanup_action_state_data:
            action_runs.append(
                cls.action_run_from_state(
                    job_run,
                    cleanup_action_state_data,
                    cleanup=True,
                ),
            )

        action_run_map = {
            maybe_decode(action_run.action_name): action_run
            for action_run in action_runs
        }
        return ActionRunCollection(job_run.action_graph, action_run_map)

    @classmethod
    def build_run_for_action(cls, job_run, action, action_runner):
        """Create an ActionRun for a JobRun and Action."""
        run_node = action.node_pool.next(
        ) if action.node_pool else job_run.node

        if action.trigger_timeout:
            trigger_timeout = job_run.run_time + action.trigger_timeout
        else:
            trigger_timeout = job_run.run_time + datetime.timedelta(days=1)

        args = {
            'job_run_id': job_run.id,
            'name': action.name,
            'node': run_node,
            'bare_command': action.command,
            'parent_context': job_run.context,
            'output_path': job_run.output_path.clone(),
            'cleanup': action.is_cleanup,
            'action_runner': action_runner,
            'retries_remaining': action.retries,
            'retries_delay': action.retries_delay,
            'executor': action.executor,
            'cpus': action.cpus,
            'mem': action.mem,
            'disk': action.disk,
            'constraints': action.constraints,
            'docker_image': action.docker_image,
            'docker_parameters': action.docker_parameters,
            'env': action.env,
            'extra_volumes': action.extra_volumes,
            'trigger_downstreams': action.trigger_downstreams,
            'triggered_by': action.triggered_by,
            'on_upstream_rerun': action.on_upstream_rerun,
            'trigger_timeout_timestamp': trigger_timeout.timestamp(),
        }
        if action.executor == ExecutorTypes.mesos.value:
            return MesosActionRun(**args)
        return SSHActionRun(**args)

    @classmethod
    def action_run_from_state(cls, job_run, state_data, cleanup=False):
        """Restore an ActionRun for this JobRun from the state data."""
        args = {
            'state_data': state_data,
            'parent_context': job_run.context,
            'output_path': job_run.output_path.clone(),
            'job_run_node': job_run.node,
            'cleanup': cleanup,
        }

        if state_data.get('executor') == ExecutorTypes.mesos.value:
            return MesosActionRun.from_state(**args)
        return SSHActionRun.from_state(**args)


class ActionRun(Observable):
    """Base class for tracking the state of a single run of an Action.

    ActionRun's state machine is observed by a parent JobRun.
    """

    CANCELLED = 'cancelled'
    FAILED = 'failed'
    QUEUED = 'queued'
    RUNNING = 'running'
    SCHEDULED = 'scheduled'
    SKIPPED = 'skipped'
    STARTING = 'starting'
    SUCCEEDED = 'succeeded'
    UNKNOWN = 'unknown'

    default_transitions = dict(fail=FAILED, success=SUCCEEDED)
    STATE_MACHINE = Machine(
        'scheduled',
        **{
            CANCELLED:
                dict(skip=SKIPPED),
            FAILED:
                dict(skip=SKIPPED),
            RUNNING:
                dict(fail_unknown=UNKNOWN, **default_transitions),
            STARTING:
                dict(started=RUNNING, fail=FAILED, fail_unknown=UNKNOWN),
            UNKNOWN:
                dict(running=RUNNING, fail_unknown=UNKNOWN, **default_transitions),
            QUEUED:
                dict(
                    cancel=CANCELLED,
                    start=STARTING,
                    schedule=SCHEDULED,
                    **default_transitions,
                ),
            SCHEDULED:
                dict(
                    ready=QUEUED,
                    queue=QUEUED,
                    cancel=CANCELLED,
                    start=STARTING,
                    **default_transitions,
                ),
        }
    )

    # The set of states that are considered end states. Technically some of
    # these states can be manually transitioned to other states.
    END_STATES = {FAILED, SUCCEEDED, CANCELLED, SKIPPED, UNKNOWN}

    # Failed render command is false to ensure that it will fail when run
    FAILED_RENDER = 'false # Command failed to render correctly. See the Tron error log.'
    NOTIFY_TRIGGER_READY = 'trigger_ready'

    EXIT_INVALID_COMMAND = -1
    EXIT_NODE_ERROR = -2
    EXIT_STOP_KILL = -3
    EXIT_TRIGGER_TIMEOUT = -4
    EXIT_MESOS_DISABLED = -5

    EXIT_REASONS = {
        EXIT_INVALID_COMMAND: 'Invalid command',
        EXIT_NODE_ERROR: 'Node error',
        EXIT_STOP_KILL: 'Stopped or killed',
        EXIT_TRIGGER_TIMEOUT: 'Timed out waiting for trigger',
        EXIT_MESOS_DISABLED: 'Mesos disabled',
    }

    context_class = command_context.ActionRunContext

    # TODO: create a class for ActionRunId, JobRunId, Etc
    def __init__(
        self,
        job_run_id,
        name,
        node,
        bare_command=None,
        parent_context=None,
        output_path=None,
        cleanup=False,
        start_time=None,
        end_time=None,
        run_state=SCHEDULED,
        rendered_command=None,
        exit_status=None,
        action_runner=None,
        retries_remaining=None,
        retries_delay=None,
        exit_statuses=None,
        machine=None,
        executor=None,
        cpus=None,
        mem=None,
        disk=None,
        constraints=None,
        docker_image=None,
        docker_parameters=None,
        env=None,
        extra_volumes=None,
        mesos_task_id=None,
        trigger_downstreams=None,
        triggered_by=None,
        on_upstream_rerun=None,
        trigger_timeout_timestamp=None,
    ):
        super().__init__()
        self.job_run_id = maybe_decode(job_run_id)
        self.action_name = maybe_decode(name)
        self.node = node
        self.start_time = start_time
        self.end_time = end_time
        self.exit_status = exit_status
        self.bare_command = maybe_decode(bare_command)
        self.rendered_command = rendered_command
        self.action_runner = action_runner or NoActionRunnerFactory()
        self.machine = machine or Machine.from_machine(
            ActionRun.STATE_MACHINE, None, run_state
        )
        self.is_cleanup = cleanup
        self.executor = executor
        self.cpus = cpus
        self.mem = mem
        self.disk = disk
        self.constraints = constraints
        self.docker_image = docker_image
        self.docker_parameters = docker_parameters
        self.env = env
        self.extra_volumes = extra_volumes
        self.mesos_task_id = mesos_task_id
        self.output_path = output_path or filehandler.OutputPath()
        self.output_path.append(self.id)
        self.context = command_context.build_context(self, parent_context)
        self.retries_remaining = retries_remaining
        self.retries_delay = retries_delay
        self.exit_statuses = exit_statuses
        self.trigger_downstreams = trigger_downstreams
        self.triggered_by = triggered_by
        self.on_upstream_rerun = on_upstream_rerun
        self.trigger_timeout_timestamp = trigger_timeout_timestamp
        self.trigger_timeout_call = None

        if self.exit_statuses is None:
            self.exit_statuses = []

        self.action_command = None
        self.in_delay = None

    @property
    def state(self):
        return self.machine.state

    @property
    def id(self):
        return f"{self.job_run_id}.{self.action_name}"

    @classmethod
    def from_state(
        cls,
        state_data,
        parent_context,
        output_path,
        job_run_node,
        cleanup=False,
    ):
        """Restore the state of this ActionRun from a serialized state."""
        pool_repo = node.NodePoolRepository.get_instance()

        # Support state from older version
        if 'id' in state_data:
            job_run_id, action_name = state_data['id'].rsplit('.', 1)
        else:
            job_run_id = state_data['job_run_id']
            action_name = state_data['action_name']

        job_run_node = pool_repo.get_node(
            state_data.get('node_name'),
            job_run_node,
        )

        action_runner_data = state_data.get('action_runner')
        if action_runner_data:
            action_runner = SubprocessActionRunnerFactory(**action_runner_data)
        else:
            action_runner = NoActionRunnerFactory()

        run = cls(
            job_run_id=job_run_id,
            name=action_name,
            node=job_run_node,
            parent_context=parent_context,
            output_path=output_path,
            rendered_command=maybe_decode(state_data.get('rendered_command')),
            bare_command=maybe_decode(state_data['command']),
            cleanup=cleanup,
            start_time=state_data['start_time'],
            end_time=state_data['end_time'],
            run_state=state_data['state'],
            exit_status=state_data.get('exit_status'),
            retries_remaining=state_data.get('retries_remaining'),
            retries_delay=state_data.get('retries_delay'),
            exit_statuses=state_data.get('exit_statuses'),
            action_runner=action_runner,
            executor=state_data.get('executor', ExecutorTypes.ssh.value),
            cpus=state_data.get('cpus'),
            mem=state_data.get('mem'),
            disk=state_data.get('disk'),
            constraints=state_data.get('constraints'),
            docker_image=state_data.get('docker_image'),
            docker_parameters=state_data.get('docker_parameters'),
            env=state_data.get('env'),
            extra_volumes=state_data.get('extra_volumes'),
            mesos_task_id=state_data.get('mesos_task_id'),
            trigger_downstreams=state_data.get('trigger_downstreams'),
            triggered_by=state_data.get('triggered_by'),
            on_upstream_rerun=state_data.get('on_upstream_rerun'),
            trigger_timeout_timestamp=state_data.get('trigger_timeout_timestamp'),
        )

        # Transition running to fail unknown because exit status was missed
        # Recovery will look for unknown runs
        if run.is_active:
            run.transition_and_notify('fail_unknown')
        return run

    def start(self):
        """Start this ActionRun."""
        if self.in_delay is not None:
            log.warning(f"{self} cancelling suspend timer")
            self.in_delay.cancel()
            self.in_delay = None

        if not self.machine.check('start'):
            return False

        if len(self.exit_statuses) == 0:
            log.info(f"{self} starting")
        else:
            log.info(f"{self} restarting, retry {len(self.exit_statuses)}")

        self.start_time = timeutils.current_time()
        self.transition_and_notify('start')

        if not self.is_valid_command:
            log.error(f"{self} invalid command: {self.bare_command}")
            self.fail(self.EXIT_INVALID_COMMAND)
            return

        return self.submit_command()

    def submit_command(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    def kill(self, final=True):
        raise NotImplementedError()

    def _done(self, target, exit_status=0):
        if self.machine.check(target):
            if self.triggered_by:
                EventBus.clear_subscriptions(self.__hash__())
            self.clear_trigger_timeout()
            self.exit_status = exit_status
            self.end_time = timeutils.current_time()
            log.info(
                f"{self} completed with {target}, transitioned to "
                f"{self.state}, exit status: {exit_status}"
            )
            return self.transition_and_notify(target)
        else:
            log.debug(
                f"{self} cannot transition from {self.state} via {target}"
            )

    def retry(self):
        """Invoked externally (via API) when action needs to be re-tried
        manually.
        """
        if self.retries_remaining is None or self.retries_remaining <= 0:
            self.retries_remaining = 1

        if self.is_done:
            return self._exit_unsuccessful(self.exit_status)
        else:
            log.info(f"{self} getting killed for a retry")
            return self.kill(final=False)

    def start_after_delay(self):
        log.info(f"{self} resuming after retry delay")
        self.machine.reset()
        self.in_delay = None
        self.start()

    def restart(self):
        """Used by `fail` when action run has to be re-tried."""
        if self.retries_delay is not None:
            self.in_delay = reactor.callLater(
                self.retries_delay.total_seconds(), self.start_after_delay
            )
            log.info(f"{self} delaying for a retry in {self.retries_delay}s")
        else:
            self.machine.reset()
            return self.start()

    def fail(self, exit_status=None):
        if self.retries_remaining:
            self.retries_remaining = -1

        return self._done('fail', exit_status)

    def _exit_unsuccessful(self, exit_status=None):
        if self.retries_remaining is not None:
            if self.retries_remaining > 0:
                self.retries_remaining -= 1
                self.exit_statuses.append(exit_status)
                return self.restart()
            else:
                log.info(
                    "Reached maximum number of retries: {}".format(
                        len(self.exit_statuses),
                    )
                )
        if exit_status is None:
            return self._done('fail_unknown', exit_status)
        else:
            return self._done('fail', exit_status)

    def triggers_to_emit(self) -> List[str]:
        if not self.trigger_downstreams:
            return []

        if isinstance(self.trigger_downstreams, bool):
            templates = ["shortdate.{shortdate}"]
        elif isinstance(self.trigger_downstreams, dict):
            templates = [
                f"{k}.{v}" for k, v in self.trigger_downstreams.items()
            ]
        else:
            log.error(f"{self} trigger_downstreams must be true or dict")

        return [self.render_template(trig) for trig in templates]

    def emit_triggers(self):
        triggers = self.triggers_to_emit()
        if not triggers:
            return

        log.info(f"{self} publishing triggers: [{', '.join(triggers)}]")
        job_id = '.'.join(self.job_run_id.split('.')[:-1])
        for trigger in triggers:
            EventBus.publish(f"{job_id}.{self.action_name}.{trigger}")

    # TODO: cache if safe
    @property
    def rendered_triggers(self) -> List[str]:
        return [self.render_template(trig) for trig in self.triggered_by or []]

    # TODO: subscribe for events and maintain a list of remaining triggers
    @property
    def remaining_triggers(self):
        return [
            trig for trig in self.rendered_triggers
            if not EventBus.has_event(trig)
        ]

    def success(self):
        if self.trigger_downstreams:
            self.emit_triggers()
        return self._done('success')

    def fail_unknown(self):
        """Failed with unknown reason."""
        log.warning(f"{self} failed with no exit code")
        return self._done('fail_unknown', None)

    def cancel_delay(self):
        if self.in_delay is not None:
            self.in_delay.cancel()
            self.in_delay = None
            self.fail(self.EXIT_STOP_KILL)
            return True

    @property
    def state_data(self):
        """This data is used to serialize the state of this action run."""
        rendered_command = self.rendered_command

        if isinstance(self.action_runner, NoActionRunnerFactory):
            action_runner = None
        else:
            action_runner = dict(
                status_path=self.action_runner.status_path,
                exec_path=self.action_runner.exec_path,
            )
        # Freeze command after it's run
        command = rendered_command if rendered_command else self.bare_command
        return {
            'job_run_id': self.job_run_id,
            'action_name': self.action_name,
            'state': self.state,
            'start_time': self.start_time,
            'end_time': self.end_time,
            'command': maybe_decode(command),
            'rendered_command': maybe_decode(self.rendered_command),
            'node_name': self.node.get_name() if self.node else None,
            'exit_status': self.exit_status,
            'retries_remaining': self.retries_remaining,
            'retries_delay': self.retries_delay,
            'exit_statuses': self.exit_statuses,
            'action_runner': action_runner,
            'executor': self.executor,
            'cpus': self.cpus,
            'mem': self.mem,
            'disk': self.disk,
            'constraints': self.constraints,
            'docker_image': self.docker_image,
            'docker_parameters': self.docker_parameters,
            'env': self.env,
            'extra_volumes': self.extra_volumes,
            'mesos_task_id': self.mesos_task_id,
            'trigger_downstreams': self.trigger_downstreams,
            'triggered_by': self.triggered_by,
            'on_upstream_rerun': self.on_upstream_rerun,
            'trigger_timeout_timestamp': self.trigger_timeout_timestamp,
        }

    def render_template(self, template):
        """Render our configured command using the command context."""
        return StringFormatter(self.context).format(template)

    def render_command(self):
        """Render our configured command using the command context."""
        return self.render_template(self.bare_command)

    @property
    def command(self):
        if self.rendered_command:
            return self.rendered_command
        try:
            self.rendered_command = self.render_command()
        except Exception as e:
            log.error(f"{self} failed rendering command: {e}")
            # Return a command string that will always fail
            self.rendered_command = self.FAILED_RENDER
        return self.rendered_command

    @property
    def is_valid_command(self):
        """Returns True if the bare_command was rendered without any errors.
        This has the side effect of actually rendering the bare_command.
        """
        return self.command != self.FAILED_RENDER

    @property
    def is_done(self):
        return self.state in self.END_STATES

    @property
    def is_complete(self):
        return self.is_succeeded or self.is_skipped

    @property
    def is_broken(self):
        return self.is_failed or self.is_cancelled or self.is_unknown

    @property
    def is_active(self):
        return self.is_starting or self.is_running

    def cleanup(self):
        self.clear_observers()
        if self.triggered_by:
            EventBus.clear_subscriptions(self.__hash__())
        self.clear_trigger_timeout()
        self.cancel()

    def clear_trigger_timeout(self):
        if self.trigger_timeout_call:
            self.trigger_timeout_call.cancel()
            self.trigger_timeout_call = None

    def setup_subscriptions(self):
        remaining_triggers = self.remaining_triggers
        if not remaining_triggers:
            return

        if self.trigger_timeout_timestamp:
            now = timeutils.current_time().timestamp()
            delay = max(self.trigger_timeout_timestamp - now, 1)
            self.trigger_timeout_call = reactor.callLater(
                delay, self.trigger_timeout_reached
            )
        else:
            log.error(f"{self} has no trigger_timeout_timestamp")

        for trigger in remaining_triggers:
            EventBus.subscribe(trigger, self.__hash__(), self.trigger_notify)

    def trigger_timeout_reached(self):
        if self.remaining_triggers:
            self.trigger_timeout_call = None
            log.warning(
                f"{self} reached timeout waiting for: {self.remaining_triggers}"
            )
            self.fail(self.EXIT_TRIGGER_TIMEOUT)
        else:
            self.notify(ActionRun.NOTIFY_TRIGGER_READY)

    def trigger_notify(self, *_):
        if not self.remaining_triggers:
            self.clear_trigger_timeout()
            self.notify(ActionRun.NOTIFY_TRIGGER_READY)

    @property
    def is_blocked_on_trigger(self):
        return not self.is_done and bool(self.remaining_triggers)

    def __getattr__(self, name: str):
        """Support convenience properties for checking if this ActionRun is in
        a specific state (Ex: self.is_running would check if self.state is
        STATE_RUNNING) or for transitioning to a new state (ex: ready).
        """
        if name in self.machine.transition_names:
            return lambda: self.transition_and_notify(name)

        if name.startswith('is_'):
            state_name = name.replace('is_', '')
            if state_name not in self.machine.states:
                raise AttributeError(f"{name} is not a state")
            return self.state == state_name
        else:
            raise AttributeError(name)

    def __str__(self):
        return f"ActionRun: {self.id}"

    def transition_and_notify(self, target):
        if self.machine.transition(target):
            self.notify(self.state)
            return True


class SSHActionRun(ActionRun, Observer):
    """An ActionRun that executes the command on a node through SSH.
    """

    def __init__(self, *args, **kwargs):
        super(SSHActionRun, self).__init__(*args, **kwargs)

    def submit_command(self):
        action_command = self.build_action_command()
        try:
            self.node.submit_command(action_command)
        except node.Error as e:
            log.warning("Failed to start %s: %r", self.id, e)
            self._exit_unsuccessful(self.EXIT_NODE_ERROR)
            return
        return True

    def stop(self):
        if self.retries_remaining is not None:
            self.retries_remaining = -1

        if self.cancel_delay():
            return

        stop_command = self.action_runner.build_stop_action_command(
            self.id,
            'terminate',
        )
        self.node.submit_command(stop_command)

    def kill(self, final=True):
        if self.retries_remaining is not None and final:
            self.retries_remaining = -1

        if self.cancel_delay():
            return

        kill_command = self.action_runner.build_stop_action_command(
            self.id,
            'kill',
        )
        self.node.submit_command(kill_command)

    def build_action_command(self):
        """Create a new ActionCommand instance to send to the node."""
        serializer = filehandler.OutputStreamSerializer(self.output_path)
        self.action_command = self.action_runner.create(
            id=self.id,
            command=self.command,
            serializer=serializer,
        )
        self.watch(self.action_command)
        return self.action_command

    def handle_action_command_state_change(self, action_command, event):
        """Observe ActionCommand state changes."""
        log.debug(
            f"{self} action_command state change: {action_command.state}"
        )

        if event == ActionCommand.RUNNING:
            return self.transition_and_notify('started')

        if event == ActionCommand.FAILSTART:
            return self._exit_unsuccessful(self.EXIT_NODE_ERROR)

        if event == ActionCommand.EXITING:
            if action_command.exit_status is None:
                return self.fail_unknown()

            if not action_command.exit_status:
                return self.success()

            return self._exit_unsuccessful(action_command.exit_status)

    handler = handle_action_command_state_change


class MesosActionRun(ActionRun, Observer):
    """An ActionRun that executes the command on a Mesos cluster.
    """
    def _create_mesos_task(self, mesos_cluster, serializer, task_id=None):
        return mesos_cluster.create_task(
            action_run_id=self.id,
            command=self.command,
            cpus=self.cpus,
            mem=self.mem,
            disk=self.disk,
            constraints=[[c.attribute, c.operator, c.value]
                         for c in self.constraints],
            docker_image=self.docker_image,
            docker_parameters=[e._asdict() for e in self.docker_parameters],
            env=self.env,
            extra_volumes=[e._asdict() for e in self.extra_volumes],
            serializer=serializer,
            task_id=task_id,
        )

    def submit_command(self):
        serializer = filehandler.OutputStreamSerializer(self.output_path)
        mesos_cluster = MesosClusterRepository.get_cluster()
        task = self._create_mesos_task(mesos_cluster, serializer)
        if not task:  # Mesos is disabled
            self.fail(self.EXIT_MESOS_DISABLED)
            return

        self.mesos_task_id = task.get_mesos_id()

        # Watch before submitting, in case submit causes a transition
        self.watch(task)
        mesos_cluster.submit(task)
        return task

    def recover(self):
        if not self.machine.check('running'):
            log.error(
                f'{self} unable to transition from {self.machine.state}'
                'to running for recovery'
            )
            return

        if self.mesos_task_id is None:
            log.error(f'{self} no task ID, cannot recover')
            self.fail_unknown()
            return

        log.info(f'{self} recovering Mesos run')

        serializer = filehandler.OutputStreamSerializer(self.output_path)
        mesos_cluster = MesosClusterRepository.get_cluster()
        task = self._create_mesos_task(
            mesos_cluster,
            serializer,
            self.mesos_task_id,
        )
        if not task:
            log.warning(
                f'{self} cannot recover, Mesos is disabled or '
                f'invalid task ID {self.mesos_task_id!r}'
            )
            self.fail_unknown()
            return

        self.watch(task)
        mesos_cluster.recover(task)

        # Reset status
        self.exit_status = None
        self.end_time = None
        self.transition_and_notify('running')

        return task

    def stop(self):
        if self.retries_remaining is not None:
            self.retries_remaining = -1

        if self.cancel_delay():
            return

        return self._kill_mesos_task()

    def kill(self, final=True):
        if self.retries_remaining is not None and final:
            self.retries_remaining = -1

        if self.cancel_delay():
            return

        return self._kill_mesos_task()

    def _kill_mesos_task(self):
        msgs = []
        if not self.is_active:
            msgs.append(
                f'Action is {self.state}, not running. Continuing anyway.'
            )

        mesos_cluster = MesosClusterRepository.get_cluster()
        if self.mesos_task_id is None:
            msgs.append("Error: Can't find task id for the action.")
        else:
            msgs.append(f"Sending kill for {self.mesos_task_id}...")
            succeeded = mesos_cluster.kill(self.mesos_task_id)
            if succeeded:
                msgs.append(
                    "Sent! It can take up to docker_stop_timeout (current setting is 2 mins) to stop."
                )
            else:
                msgs.append(
                    "Error while sending kill request. Please try again."
                )

        return '\n'.join(msgs)

    def handle_action_command_state_change(self, action_command, event):
        """Observe ActionCommand state changes."""
        log.debug(
            f"{self} action_command state change: {action_command.state}"
        )

        if event == ActionCommand.RUNNING:
            return self.transition_and_notify('started')

        if event == ActionCommand.FAILSTART:
            return self._exit_unsuccessful(action_command.exit_status)

        if event == ActionCommand.EXITING:
            if action_command.exit_status is None:
                # This is different from SSHActionRun
                # Allows retries to happen, if configured
                return self._exit_unsuccessful(None)

            if not action_command.exit_status:
                return self.success()

            return self._exit_unsuccessful(action_command.exit_status)

    handler = handle_action_command_state_change


def min_filter(seq):
    seq = list(filter(None, seq))
    return min(seq) if any(seq) else None


def eager_all(seq):
    return all(list(seq))


class ActionRunCollection(object):
    """A collection of ActionRuns used by a JobRun."""

    def __init__(self, action_graph, run_map):
        self.action_graph = action_graph
        self.run_map = run_map
        # Setup proxies
        self.proxy_action_runs_with_cleanup = proxy.CollectionProxy(
            self.get_action_runs_with_cleanup,
            [
                proxy.attr_proxy('is_running', any),
                proxy.attr_proxy('is_starting', any),
                proxy.attr_proxy('is_scheduled', any),
                proxy.attr_proxy('is_cancelled', any),
                proxy.attr_proxy('is_active', any),
                proxy.attr_proxy('is_queued', all),
                proxy.attr_proxy('is_complete', all),
                proxy.func_proxy('queue', eager_all),
                proxy.func_proxy('cancel', eager_all),
                proxy.func_proxy('success', eager_all),
                proxy.func_proxy('fail', eager_all),
                proxy.func_proxy('ready', eager_all),
                proxy.func_proxy('cleanup', eager_all),
                proxy.func_proxy('stop', eager_all),
                proxy.attr_proxy('start_time', min_filter),
                proxy.attr_proxy('state_data', eager_all),
            ],
        )

    def action_runs_for_actions(self, actions):
        return (
            self.run_map[a.name] for a in actions if a.name in self.run_map
        )

    def get_action_runs_with_cleanup(self):
        return self.run_map.values()

    action_runs_with_cleanup = property(get_action_runs_with_cleanup)

    def get_action_runs(self):
        return (run for run in self.run_map.values() if not run.is_cleanup)

    action_runs = property(get_action_runs)

    @property
    def cleanup_action_run(self) -> ActionRun:
        return self.run_map.get(action.CLEANUP_ACTION_NAME)

    @property
    def state_data(self):
        return [run.state_data for run in self.action_runs]

    @property
    def cleanup_action_state_data(self):
        if self.cleanup_action_run:
            return self.cleanup_action_run.state_data

    def get_startable_action_runs(self):
        """Returns any actions that are scheduled or queued that can be run."""

        return [
            r for r in self.action_runs
            if r.machine.check('start') and not self._is_run_blocked(r)
        ]

    @property
    def has_startable_action_runs(self):
        return any(self.get_startable_action_runs())

    def _is_run_blocked(self, action_run):
        """Returns True if the ActionRun is waiting on a required run to
        finish before it can run.
        """
        if action_run.is_done or action_run.is_active:
            return False

        required_actions = self.action_graph.get_required_actions(
            action_run.action_name,
        )

        if required_actions:
            required_runs = self.action_runs_for_actions(required_actions)
            if any(not run.is_complete for run in required_runs):
                return True

        if action_run.is_blocked_on_trigger:
            return True

        return False

    @property
    def is_blocked_on_trigger(self):
        return any(r.is_blocked_on_trigger for r in self.action_runs)

    @property
    def is_done(self):
        """Returns True when there are no running ActionRuns and all
        non-blocked ActionRuns are done.
        """
        if self.is_running:
            return False

        def done_or_blocked(action_run):
            return action_run.is_done or self._is_run_blocked(action_run)

        return all(done_or_blocked(run) for run in self.action_runs)

    @property
    def is_failed(self):
        """Return True if there are failed actions and all ActionRuns are
        done or blocked.
        """
        return self.is_done and any(run.is_failed for run in self.action_runs)

    @property
    def is_complete_without_cleanup(self):
        return all(run.is_complete for run in self.action_runs)

    @property
    def names(self):
        return self.run_map.keys()

    @property
    def end_time(self):
        if not self.is_done:
            return None
        end_times = list(
            run.end_time
            for run in self.get_action_runs_with_cleanup()
            if run.end_time
        )
        return max(end_times) if any(end_times) else None

    def __str__(self):
        def blocked_state(action_run):
            return ":blocked" if self._is_run_blocked(action_run) else ""

        run_states = ', '.join(
            f"{a.action_name}({a.state}{blocked_state(a)})"
            for a in self.run_map.values()
        )
        return f"{self.__class__.__name__}[{run_states}]"

    def __getattr__(self, name):
        return self.proxy_action_runs_with_cleanup.perform(name)

    def __getitem__(self, name):
        return self.run_map[name]

    def __contains__(self, name):
        return name in self.run_map

    def __iter__(self):
        return iter(self.run_map.values())

    def get(self, name):
        return self.run_map.get(name)



import mock

from testifycompat import assert_equal
from testifycompat import assert_in
from testifycompat import assert_not_in
from testifycompat import assert_raises
from testifycompat import setup
from testifycompat import TestCase
from tests.assertions import assert_mock_calls
from tests.testingutils import autospec_method
from tron.utils import collections


class TestMappingCollections(TestCase):
    @setup
    def setup_collection(self):
        self.name = 'some_name'
        self.collection = collections.MappingCollection(self.name)

    def test_filter_by_name(self):
        autospec_method(self.collection.remove)
        self.collection.update(dict.fromkeys(['c', 'd', 'e']))
        self.collection.filter_by_name(['a', 'c'])
        expected = [mock.call(name) for name in ['d', 'e']]
        assert_mock_calls(expected, self.collection.remove.mock_calls)

    def test_remove_missing(self):
        assert_raises(ValueError, self.collection.remove, 'name')

    def test_remove(self):
        name = 'the_name'
        self.collection[name] = item = mock.Mock()
        self.collection.remove(name)
        assert_not_in(name, self.collection)
        item.disable.assert_called_with()

    def test_contains_item_false(self):
        mock_item, mock_func = mock.Mock(), mock.Mock()
        assert not self.collection.contains_item(mock_item, mock_func)
        assert not mock_func.mock_calls

    def test_contains_item_not_equal(self):
        mock_item, mock_func = mock.Mock(), mock.Mock()
        self.collection[mock_item.get_name()] = 'other item'
        result = self.collection.contains_item(mock_item, mock_func)
        assert_equal(result, mock_func.return_value)
        mock_func.assert_called_with(mock_item)

    def test_contains_item_true(self):
        mock_item, mock_func = mock.Mock(), mock.Mock()
        self.collection[mock_item.get_name()] = mock_item
        assert self.collection.contains_item(mock_item, mock_func)

    def test_add_contains(self):
        autospec_method(self.collection.contains_item)
        item, update_func = mock.Mock(), mock.Mock()
        assert not self.collection.add(item, update_func)
        assert_not_in(item.get_name(), self.collection)

    def test_add_new(self):
        autospec_method(self.collection.contains_item, return_value=False)
        item, update_func = mock.Mock(), mock.Mock()
        assert self.collection.add(item, update_func)
        assert_in(item.get_name(), self.collection)

    def test_replace(self):
        autospec_method(self.collection.add)
        item = mock.Mock()
        self.collection.replace(item)
        self.collection.add.assert_called_with(
            item,
            self.collection.remove_item,
        )



from __future__ import absolute_import
from __future__ import unicode_literals

__version_info__ = (0, 9, 9, 11)
__version__ = ".".join("%s" % v for v in __version_info__)
__author__ = 'Yelp <yelplabs@yelp.com>'
__credits__ = [
    'Rhett Garber <rhettg@gmail.com>',
    'Eugene Baumstein <eugene.baumstein@gmail>',
    'James Brown <jbrown@yelp.com>',
    'Adam Derewecki <derewecki@gmail.com>',
    'Steve Johnson <sjohnson@yelp.com>',
    'Daniel Nephin <dnephin@yelp.com>',
    'Matthew Tytel <matthewtytel@gmail.com>',
]



#!/usr/bin/env python
"""

.. warning::

    This script is still very experimental. Use at your own risk. It will
    be replaced over time with lib2to3 fixers.


Usage:

    ``python -m testifycompat.bin.migrate <filenames>``

Example:

    ``find tests -name *.py | xargs python migrate.py``


"""
import functools
import re
import sys


def replace(pattern, repl):
    return functools.partial(re.sub, pattern, repl)


replaces = [
    # Replace imports
    replace(r'^from testify import ', 'from testifycompat import '),
    replace(r'^from testify.assertions import ', 'from testifycompat import '),
    replace(r'^import testify as T', 'import testifycompat as T'),

    # Replace test classes
    replace(r'^class (?:Test)?(\w+)(?:Test|TestCase)\((?:T\.)?TestCase\):$',
            'class Test\\1(object):'),
    replace(r'^class (?:Test)?(\w+)(?:Test|TestCase)(\(\w+TestCase\)):$',
            'class Test\\1\\2:'),

    # Replace some old assertions
    replace(r'self.assert_\((.*)\)', 'assert \\1')
]


def run_replacement(contents):
    for line in contents:
        for replacement in replaces:
            line = replacement(line)
        yield line


def strip_if_main_run(contents):
    if len(contents) < 2:
        return contents
    if 'run()' in contents[-1] and 'if __name__ == ' in contents[-2]:
        return contents[:-2]
    return contents


def run_migration_on_file(filename):
    with open(filename, 'r') as fh:
        lines = fh.read().split('\n')

    lines = list(run_replacement(lines))
    lines = strip_if_main_run(lines)

    with open(filename, 'w') as fh:
        fh.write('\n'.join(lines))


if __name__ == "__main__":
    for filename in sys.argv[1:]:
        run_migration_on_file(filename)



import logging

from tron import command_context
from tron import node
from tron.core import actiongraph
from tron.core import jobrun
from tron.core.actionrun import ActionRun
from tron.serialize import filehandler
from tron.utils import maybe_decode
from tron.utils.observer import Observable
from tron.utils.observer import Observer


class Error(Exception):
    pass


class ConfigBuildMismatchError(Error):
    pass


class InvalidStartStateError(Error):
    pass


log = logging.getLogger(__name__)


class Job(Observable, Observer):
    """A configurable data object.

    Job uses JobRunCollection to manage its runs, and ActionGraph to manage its
    actions and their dependency graph.
    """

    STATUS_DISABLED = "disabled"
    STATUS_ENABLED = "enabled"
    STATUS_UNKNOWN = "unknown"
    STATUS_RUNNING = "running"

    NOTIFY_STATE_CHANGE = 'notify_state_change'
    NOTIFY_RUN_DONE = 'notify_run_done'

    context_class = command_context.JobContext

    # These attributes determine equality between two Job objects
    equality_attributes = [
        'name',
        'queueing',
        'scheduler',
        'node_pool',
        'all_nodes',
        'action_graph',
        'output_path',
        'action_runner',
        'max_runtime',
        'allow_overlap',
        'monitoring',
        'time_zone',
        'expected_runtime',
    ]

    # TODO: use config object
    def __init__(
        self,
        name,
        scheduler,
        queueing=True,
        all_nodes=False,
        monitoring=None,
        node_pool=None,
        enabled=True,
        action_graph=None,
        run_collection=None,
        parent_context=None,
        output_path=None,
        allow_overlap=None,
        action_runner=None,
        max_runtime=None,
        time_zone=None,
        expected_runtime=None
    ):
        super(Job, self).__init__()
        self.name = maybe_decode(name)
        self.monitoring = monitoring
        self.action_graph = action_graph
        self.scheduler = scheduler
        self.runs = run_collection
        self.queueing = queueing
        self.all_nodes = all_nodes
        self.enabled = enabled  # current enabled setting
        self.config_enabled = enabled  # enabled attribute from file
        self.node_pool = node_pool
        self.allow_overlap = allow_overlap
        self.action_runner = action_runner
        self.max_runtime = max_runtime
        self.time_zone = time_zone
        self.expected_runtime = expected_runtime
        self.output_path = output_path or filehandler.OutputPath()
        self.output_path.append(name)
        self.context = command_context.build_context(self, parent_context)
        log.info(f'{self} created')

    @classmethod
    def from_config(
        cls,
        job_config,
        scheduler,
        parent_context,
        output_path,
        action_runner,
    ):
        """Factory method to create a new Job instance from configuration."""
        action_graph = actiongraph.ActionGraph.from_config(
            job_config.actions,
            job_config.cleanup_action,
        )
        runs = jobrun.JobRunCollection.from_config(job_config)
        node_repo = node.NodePoolRepository.get_instance()

        return cls(
            name=job_config.name,
            monitoring=job_config.monitoring,
            time_zone=job_config.time_zone,
            queueing=job_config.queueing,
            all_nodes=job_config.all_nodes,
            node_pool=node_repo.get_by_name(job_config.node),
            scheduler=scheduler,
            enabled=job_config.enabled,
            run_collection=runs,
            action_graph=action_graph,
            parent_context=parent_context,
            output_path=output_path,
            allow_overlap=job_config.allow_overlap,
            action_runner=action_runner,
            max_runtime=job_config.max_runtime,
            expected_runtime=job_config.expected_runtime,
        )

    def update_from_job(self, job):
        """Update this Jobs configuration from a new config. This method
        actually takes an already constructed job and copies out its
        configuration data.
        """
        for attr in self.equality_attributes:
            setattr(self, attr, getattr(job, attr))
        log.info(f'{self} reconfigured')

    @property
    def status(self):
        """Current status."""
        if not self.enabled:
            return self.STATUS_DISABLED
        if self.runs.get_run_by_state(ActionRun.RUNNING):
            return self.STATUS_RUNNING

        if self.runs.get_run_by_state(ActionRun.SCHEDULED):
            return self.STATUS_ENABLED

        log.warning("%s in an unknown state: %s" % (self, self.runs))
        return self.STATUS_UNKNOWN

    def get_name(self):
        return self.name

    def get_monitoring(self):
        return self.monitoring

    def get_time_zone(self):
        return self.time_zone

    def get_runs(self):
        return self.runs

    @property
    def state_data(self):
        """This data is used to serialize the state of this job."""
        return {
            'runs': self.runs.state_data,
            'enabled': self.enabled,
        }

    def get_job_runs_from_state(self, state_data):
        """Apply a previous state to this Job."""
        self.enabled = state_data['enabled']
        job_runs = jobrun.job_runs_from_state(
            state_data['runs'],
            self.action_graph,
            self.output_path.clone(),
            self.context,
            self.node_pool,
        )
        return job_runs

    def build_new_runs(self, run_time, manual=False):
        """Uses its JobCollection to build new JobRuns. If all_nodes is set,
        build a run for every node, otherwise just builds a single run on a
        single node.
        """
        pool = self.node_pool
        nodes = pool.nodes if self.all_nodes else [pool.next()]
        for n in nodes:
            run = self.runs.build_new_run(self, run_time, n, manual=manual)
            self.watch(run)
            yield run

    def handle_job_run_state_change(self, _job_run, event):
        """Handle state changes from JobRuns and propagate changes to any
        observers.
        """
        # Propagate state change for serialization
        if event == jobrun.JobRun.NOTIFY_STATE_CHANGED:
            self.notify(self.NOTIFY_STATE_CHANGE)
            return

        # Propagate DONE JobRun notifications to JobScheduler
        if event == jobrun.JobRun.NOTIFY_DONE:
            self.notify(self.NOTIFY_RUN_DONE)
            return

    handler = handle_job_run_state_change

    def __eq__(self, other):
        return all(
            getattr(other, attr, None) == getattr(self, attr, None)
            for attr in self.equality_attributes
        )

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return "Job:%s" % self.name



"""Functions for returning validated values from a twisted.web.Request object.
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


def get_integer(request, key):
    """Returns the first value in the request args for the given key, if that
    value is an integer. Otherwise returns None.
    """
    value = get_string(request, key)
    if value is None or not value.isdigit():
        return None

    return int(value)


def get_string(request, key):
    """Returns the first value in the request args for a given key."""
    if not request.args:
        return None

    if type(key) is not bytes:
        key = key.encode()

    if key not in request.args:
        return None

    val = request.args[key][0]
    if val is not None and type(val) is bytes:
        val = val.decode()

    return val


def get_bool(request, key, default=None):
    """Returns True if the key exists and is truthy in the request args."""
    int_value = get_integer(request, key)
    if int_value is None:
        return default

    return bool(int_value)


def get_datetime(request, key):
    """Returns the first value in the request args for a given key. Casts to
    a datetime. Returns None if the value cannot be converted to datetime.
    """
    val = get_string(request, key)
    if not val:
        return None

    try:
        return datetime.datetime.strptime(val, DATE_FORMAT)
    except ValueError:
        return None



from __future__ import absolute_import
from __future__ import unicode_literals

from unittest import mock

from testifycompat import assert_equal
from testifycompat import assert_raises
from testifycompat import run
from testifycompat import setup
from testifycompat import TestCase
from tron.core import actiongraph


class TestActionGraph(TestCase):
    @setup
    def setup_graph(self):
        self.action_names = [
            'base_one',
            'base_two',
            'dep_one',
            'dep_one_one',
            'dep_multi',
        ]
        am = self.action_map = {}
        for name in self.action_names:
            am[name] = mock.MagicMock()
            am[name].name = name

        am['dep_multi'].required_actions = [am['dep_one_one'], am['base_two']]
        am['dep_one_one'].required_actions = [am['dep_one']]
        am['dep_one'].required_actions = [am['base_one']]

        self.graph = [am['base_one'], am['base_two']]
        self.action_graph = actiongraph.ActionGraph(self.graph, am)

    def test_from_config(self):
        config = {}
        for name in self.action_names:
            config[name] = mock.MagicMock(name=name, node='first', requires=[])
            config[name].name = name
            config[name].node = 'first'
            config[name].requires = []
        config['dep_multi'].requires = ['dep_one_one', 'base_two']
        config['dep_one_one'].requires = ['dep_one']
        config['dep_one'].requires = ['base_one']

        built_graph = actiongraph.ActionGraph.from_config(config)
        am = built_graph.action_map

        graph_base_names = {a.name for a in built_graph.graph}
        assert_equal(graph_base_names, {a.name for a in self.graph})
        assert_equal(graph_base_names, {'base_one', 'base_two'})
        assert_equal(
            set(am['dep_multi'].required_actions),
            {'dep_one_one', 'base_two'},
        )

        assert_equal(set(am.keys()), set(self.action_names))
        assert_equal(am['base_one'].dependent_actions, {'dep_one'})
        assert_equal(am['dep_one'].dependent_actions, {'dep_one_one'})

    def test_actions_for_names(self):
        actions = list(
            self.action_graph.actions_for_names(['base_one', 'dep_multi']),
        )
        expected_actions = [
            self.action_map['base_one'],
            self.action_map['dep_multi'],
        ]
        assert_equal(actions, expected_actions)

    def test__getitem__(self):
        assert_equal(
            self.action_graph['base_one'],
            self.action_map['base_one'],
        )

    def test__getitem__miss(self):
        assert_raises(KeyError, lambda: self.action_graph['unknown'])

    def test__eq__(self):
        other_graph = mock.MagicMock(
            graph=self.graph,
            action_map=self.action_map,
        )
        assert_equal(self.action_graph, other_graph)

        other_graph.graph = None
        assert not self.action_graph == other_graph

    def test__ne__(self):
        other_graph = mock.MagicMock()
        assert self.graph != other_graph


if __name__ == "__main__":
    run()



from pyformance.meters import Counter
from pyformance.meters import Histogram
from pyformance.meters import Meter
from pyformance.meters import SimpleGauge
from pyformance.meters import Timer

all_metrics = {}


def get_metric(metric_type, name, dimensions, default):
    global all_metrics
    dimensions = tuple(sorted(dimensions.items())) if dimensions else ()
    key = (metric_type, name, dimensions)
    return all_metrics.setdefault(key, default)


def timer(name, delta, dimensions=None):
    timer = get_metric('timer', name, dimensions, Timer())
    timer._update(delta)


def count(name, inc=1, dimensions=None):
    counter = get_metric('counter', name, dimensions, Counter())
    counter.inc(inc)


def meter(name, dimensions=None):
    meter = get_metric('meter', name, dimensions, Meter())
    meter.mark()


def gauge(name, value, dimensions=None):
    gauge = get_metric('gauge', name, dimensions, SimpleGauge())
    gauge.set_value(value)


def histogram(name, value, dimensions=None):
    histogram = get_metric('histogram', name, dimensions, Histogram())
    histogram.add(value)


def view_timer(timer):
    data = view_meter(timer)
    data.update(view_histogram(timer))
    return data


def view_counter(counter):
    return {'count': counter.get_count()}


def view_meter(meter):
    return {
        'count': meter.get_count(),
        'm1_rate': meter.get_one_minute_rate(),
        'm5_rate': meter.get_five_minute_rate(),
        'm15_rate': meter.get_fifteen_minute_rate(),
    }


def view_gauge(gauge):
    return {'value': gauge.get_value()}


def view_histogram(histogram):
    snapshot = histogram.get_snapshot()
    return {
        'count': histogram.get_count(),
        'mean': histogram.get_mean(),
        'min': histogram.get_min(),
        'max': histogram.get_max(),
        'p50': snapshot.get_median(),
        'p75': snapshot.get_75th_percentile(),
        'p95': snapshot.get_95th_percentile(),
        'p99': snapshot.get_99th_percentile(),
    }


metrics_to_viewers = {
    'counter': view_counter,
    'gauge': view_gauge,
    'histogram': view_histogram,
    'meter': view_meter,
    'timer': view_timer,
}


def view_all_metrics():
    all_data = {metric_type: [] for metric_type in metrics_to_viewers}
    for (metric_type, name, dims), metric in all_metrics.items():
        data = {'name': name, **metrics_to_viewers[metric_type](metric)}
        if dims:
            data.update({'dimensions': dict(dims)})
        all_data[metric_type].append(data)
    return all_data



from __future__ import absolute_import
from __future__ import unicode_literals

import os
import tempfile

import mock

from testifycompat import setup
from testifycompat import teardown
from testifycompat import TestCase
from tron.trondaemon import TronDaemon


class TronDaemonTestCase(TestCase):
    @setup
    @mock.patch('tron.trondaemon.setup_logging', mock.Mock(), autospec=None)
    @mock.patch('signal.signal', mock.Mock(), autospec=None)
    def setup(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        trond_opts = mock.Mock()
        trond_opts.working_dir = self.tmpdir.name
        trond_opts.lock_file = os.path.join(self.tmpdir.name, "lockfile")
        self.trond = TronDaemon(trond_opts)

    @teardown
    def teardown(self):
        self.tmpdir.cleanup()

    @mock.patch('tron.trondaemon.setup_logging', mock.Mock(), autospec=None)
    @mock.patch('signal.signal', mock.Mock(), autospec=None)
    def test_init(self):
        daemon = TronDaemon.__new__(TronDaemon)  # skip __init__
        options = mock.Mock()

        with mock.patch(
            'tron.utils.flock',
            autospec=True,
        ) as mock_flock:
            daemon.__init__(options)
            assert mock_flock.call_count == 0

    def test_run_uses_context(self):
        with mock.patch(
            'tron.trondaemon.setup_logging', mock.Mock(), autospec=None
        ), mock.patch(
            'tron.trondaemon.no_daemon_context', mock.Mock(), autospec=None
        ) as ndc:
            ndc.return_value = mock.MagicMock()
            ndc.return_value.__enter__.side_effect = RuntimeError()
            daemon = TronDaemon(mock.Mock())
            try:
                daemon.run()
            except Exception:
                pass
            assert ndc.call_count == 1

    def test_run_manhole_new_manhole(self):
        with open(self.trond.manhole_sock, 'w+'):
            pass

        with mock.patch(
            'twisted.internet.reactor.listenUNIX',
            autospec=True,
        ) as mock_listenUNIX:
            self.trond._run_manhole()

            assert mock_listenUNIX.call_count == 1
            # _run_manhole will remove the old manhole.sock but not recreate
            # it because we mocked out listenUNIX
            assert not os.path.exists(self.trond.manhole_sock)



"""
Parse and validate scheduler configuration and return immutable structures.
"""
import calendar
import datetime
import re
from collections import namedtuple

from tron.config import config_utils
from tron.config import ConfigError
from tron.config import schema
from tron.utils import crontab

ConfigGenericSchedule = schema.config_object_factory(
    'ConfigGenericSchedule',
    ['type', 'value'],
    ['jitter'],
)

ConfigGrocScheduler = namedtuple(
    'ConfigGrocScheduler',
    'original ordinals weekdays monthdays months timestr jitter',
)

ConfigCronScheduler = namedtuple(
    'ConfigCronScheduler',
    'original minutes hours monthdays months weekdays ordinals jitter',
)

ConfigDailyScheduler = namedtuple(
    'ConfigDailyScheduler',
    'original hour minute second days jitter',
)

ConfigConstantScheduler = namedtuple('ConfigConstantScheduler', [])

ConfigIntervalScheduler = namedtuple(
    'ConfigIntervalScheduler',
    'timedelta jitter',
)


class ScheduleParseError(ConfigError):
    pass


def pad_sequence(seq, size, padding=None):
    """Force a sequence to size. Pad with padding if too short, and ignore
    extra pieces if too long."""
    return (list(seq) + [padding for _ in range(size)])[:size]


def schedule_config_from_string(schedule, config_context):
    """Return a scheduler config object from a string."""
    schedule = schedule.strip()
    name, schedule_config = pad_sequence(
        schedule.split(None, 1),
        2,
        padding='',
    )
    if name not in schedulers:
        config = ConfigGenericSchedule('groc daily', schedule, jitter=None)
        return parse_groc_expression(config, config_context)

    config = ConfigGenericSchedule(name, schedule_config, jitter=None)
    return validate_generic_schedule_config(config, config_context)


def validate_generic_schedule_config(config, config_context):
    return schedulers[config.type](config, config_context)


# TODO: remove in 0.7
def schedule_config_from_legacy_dict(schedule, config_context):
    """Support old style schedules as dicts."""
    if 'interval' in schedule:
        config = ConfigGenericSchedule('interval', schedule['interval'], None)
        return valid_interval_scheduler(config, config_context)

    if 'start_time' in schedule or 'days' in schedule:
        start_time = schedule.get('start_time', '00:00:00')
        days = schedule.get('days', '')
        scheduler_config = '%s %s' % (start_time, days)
        config = ConfigGenericSchedule('daily', scheduler_config, None)
        return valid_daily_scheduler(config, config_context)

    path = config_context.path
    raise ConfigError("Unknown scheduler at %s: %s" % (path, schedule))


def valid_schedule(schedule, config_context):
    if isinstance(schedule, str):
        return schedule_config_from_string(schedule, config_context)

    if 'type' not in schedule:
        return schedule_config_from_legacy_dict(schedule, config_context)

    schedule = ScheduleValidator().validate(schedule, config_context)
    return validate_generic_schedule_config(schedule, config_context)


def valid_constant_scheduler(_config, _context):
    """Adapter for validation interface and constant scheduler."""
    return ConfigConstantScheduler()


def valid_daily_scheduler(config, config_context):
    """Daily scheduler, accepts a time of day and an optional list of days."""
    schedule_config = config.value
    time_string, days = pad_sequence(schedule_config.split(), 2)
    time_string = time_string or '00:00:00'
    time_spec = config_utils.valid_time(time_string, config_context)
    days = config_utils.valid_string(days or "", config_context)

    def valid_day(day):
        if day not in CONVERT_DAYS_INT:
            raise ConfigError(
                "Unknown day %s at %s" % (
                    day,
                    config_context.path,
                )
            )
        return CONVERT_DAYS_INT[day]

    original = "%s %s" % (time_string, days)
    weekdays = {valid_day(day) for day in days or ()}
    return ConfigDailyScheduler(
        original,
        time_spec.hour,
        time_spec.minute,
        time_spec.second,
        weekdays,
        jitter=config.jitter,
    )


# Shortcut values for intervals
TIME_INTERVAL_SHORTCUTS = {
    'hourly': datetime.timedelta(hours=1),
}


def valid_interval_scheduler(config, config_context):
    def build_config(delta):
        return ConfigIntervalScheduler(timedelta=delta, jitter=config.jitter)

    interval_key = config.value.strip()
    if interval_key in TIME_INTERVAL_SHORTCUTS:
        return build_config(TIME_INTERVAL_SHORTCUTS[interval_key])

    return build_config(
        config_utils.valid_time_delta(config.value, config_context),
    )


def normalize_weekdays(seq):
    return seq[6:7] + seq[:6]


def day_canonicalization_map():
    """Build a map of weekday synonym to int index 0-6 inclusive."""
    canon_map = dict()

    # 7-element lists with weekday names in order
    weekday_lists = [
        normalize_weekdays(calendar.day_name),
        normalize_weekdays(calendar.day_abbr),
        (
            'u',
            'm',
            't',
            'w',
            'r',
            'f',
            's',
        ),
        (
            'su',
            'mo',
            'tu',
            'we',
            'th',
            'fr',
            'sa',
        ),
    ]
    for day_list in weekday_lists:
        for day_name_synonym, day_index in zip(day_list, range(7)):
            canon_map[day_name_synonym] = day_index
            canon_map[day_name_synonym.lower()] = day_index
            canon_map[day_name_synonym.upper()] = day_index

    return canon_map


# Canonicalize weekday names to integer indices
CONVERT_DAYS_INT = day_canonicalization_map()  # day name/abbrev => {0123456}


def month_canonicalization_map():
    """Build a map of month synonym to int index 0-11 inclusive."""
    canon_map = dict()

    # calendar stores month data with a useless element in front. cut it off.
    monthname_lists = (calendar.month_name[1:], calendar.month_abbr[1:])
    for month_list in monthname_lists:
        for key, value in zip(month_list, range(1, 13)):
            canon_map[key] = value
            canon_map[key.lower()] = value
    return canon_map


# Canonicalize month names to integer indices
# month name/abbrev => {0 <= k <= 11}
CONVERT_MONTHS = month_canonicalization_map()


def build_groc_schedule_parser_re():
    """Build a regular expression that matches this:

        ("every"|ordinal) (day) ["of|in" (monthspec)] (["at"] HH:MM)

    ordinal   - comma-separated list of "1st" and so forth
    days      - comma-separated list of days of the week (for example,
                "mon", "tuesday", with both short and long forms being
                accepted); "every day" is equivalent to
                "every mon,tue,wed,thu,fri,sat,sun"
    monthspec - comma-separated list of month names (for example,
                "jan", "march", "sep"). If omitted, implies every month.
                You can also say "month" to mean every month, as in
                "1,8th,15,22nd of month 09:00".
    HH:MM     - time of day in 24 hour time.

    This is a slightly more permissive version of Google App Engine's schedule
    parser, documented here:
    http://code.google.com/appengine/docs/python/config/cron.html#The_Schedule_Format
    """

    # m|mon|monday|...|day
    DAY_VALUES = '|'.join(list(CONVERT_DAYS_INT.keys()) + ['day'])

    # jan|january|...|month
    MONTH_VALUES = '|'.join(list(CONVERT_MONTHS.keys()) + ['month'])

    DATE_SUFFIXES = 'st|nd|rd|th'

    # every|1st|2nd|3rd (also would accept 3nd, 1rd, 4st)
    MONTH_DAYS_EXPR = r'(?P<month_days>every|((\d+(%s),?)+))?' % DATE_SUFFIXES
    DAYS_EXPR = r'((?P<days>((%s),?)+))?' % DAY_VALUES
    MONTHS_EXPR = r'((in|of)\s+(?P<months>((%s),?)+))?' % MONTH_VALUES

    # [at] 00:00
    TIME_EXPR = r'((at\s+)?(?P<time>\d\d:\d\d))?'

    DAILY_SCHEDULE_EXPR = ''.join([
        r'^',
        MONTH_DAYS_EXPR,
        r'\s*',
        DAYS_EXPR,
        r'\s*',
        MONTHS_EXPR,
        r'\s*',
        TIME_EXPR,
        r'\s*',
        r'$',
    ])
    return re.compile(DAILY_SCHEDULE_EXPR)


# Matches expressions of the form
# ``("every"|ordinal) (days) ["of|in" (monthspec)] (["at"] HH:MM)``.
# See :py:func:`daily_schedule_parser_re` for details.
DAILY_SCHEDULE_RE = build_groc_schedule_parser_re()


def _parse_number(day):
    return int(''.join(c for c in day if c.isdigit()))


def parse_groc_expression(config, config_context):
    """Given an expression of the form in the docstring of
    daily_schedule_parser_re(), return the parsed values in a
    ConfigGrocScheduler
    """
    expression = config.value
    m = DAILY_SCHEDULE_RE.match(expression.lower())
    if not m:
        msg = 'Schedule at %s is not a valid expression: %s'
        raise ScheduleParseError(msg % (config_context.path, expression))

    timestr = m.group('time')
    if timestr is None:
        timestr = '00:00'

    if m.group('days') in (None, 'day'):
        weekdays = None
    else:
        weekdays = {CONVERT_DAYS_INT[d] for d in m.group('days').split(',')}

    monthdays = None
    ordinals = None

    if m.group('month_days') != 'every':
        values = {_parse_number(n) for n in m.group('month_days').split(',')}
        if weekdays is None:
            monthdays = values
        else:
            ordinals = values

    if m.group('months') in (None, 'month'):
        months = None
    else:
        months = {CONVERT_MONTHS[mo] for mo in m.group('months').split(',')}

    return ConfigGrocScheduler(
        original=expression,
        ordinals=ordinals,
        weekdays=weekdays,
        monthdays=monthdays,
        months=months,
        timestr=timestr,
        jitter=config.jitter,
    )


def valid_cron_scheduler(config, config_context):
    """Parse a cron schedule."""
    try:
        crontab_kwargs = crontab.parse_crontab(config.value)
        return ConfigCronScheduler(
            original=config.value, jitter=config.jitter, **crontab_kwargs
        )
    except ValueError as e:
        msg = "Invalid cron scheduler %s: %s"
        raise ConfigError(msg % (config_context.path, e))


schedulers = {
    'constant': valid_constant_scheduler,
    'daily': valid_daily_scheduler,
    'interval': valid_interval_scheduler,
    'cron': valid_cron_scheduler,
    'groc daily': parse_groc_expression,
}


class ScheduleValidator(config_utils.Validator):
    """Validate the structure of a scheduler config."""
    config_class = ConfigGenericSchedule
    defaults = {
        'jitter': datetime.timedelta(),
    }
    validators = {
        'type': config_utils.build_enum_validator(schedulers.keys()),
        'jitter': config_utils.valid_time_delta,
    }



import threading
import time

from twisted.internet import threads
from twisted.web import server

from tron.metrics import timer


def report_resource_request(resource, request, duration_ms):
    timer(
        name=f'tron.api.{resource.__class__.__name__}',
        delta=duration_ms,
        dimensions={'method': request.method.decode()}
    )


class AsyncResource():
    capacity = 10
    semaphore = threading.Semaphore(value=capacity)
    lock = threading.Lock()

    @staticmethod
    def finish(result, request, resource):
        result, duration_ms = result
        request.write(result)
        request.finish()
        report_resource_request(resource, request, duration_ms)

    @staticmethod
    def process(fn, resource, request):
        start = time.time()
        with AsyncResource.semaphore:
            result = fn(resource, request)
        duration_ms = 1000 * (time.time() - start)
        return result, duration_ms

    @staticmethod
    def bounded(fn):
        def wrapper(resource, request):
            d = threads.deferToThread(
                AsyncResource.process, fn, resource, request
            )
            d.addCallback(AsyncResource.finish, request, resource)
            d.addErrback(request.processingFailed)
            return server.NOT_DONE_YET

        return wrapper

    @staticmethod
    def exclusive(fn):
        def wrapper(resource, request):
            # ensures only one exclusive request starts consuming the semaphore
            start = time.time()
            with AsyncResource.lock:
                # this will wait until all bounded requests finished processing
                for _ in range(AsyncResource.capacity):
                    AsyncResource.semaphore.acquire()
                try:
                    return fn(resource, request)
                finally:
                    for _ in range(AsyncResource.capacity):
                        AsyncResource.semaphore.release()
                    duration_ms = 1000 * (time.time() - start)
                    report_resource_request(resource, request, duration_ms)

        return wrapper



from __future__ import absolute_import
from __future__ import unicode_literals

import shutil
import tempfile

import mock

from testifycompat import assert_equal
from testifycompat import run
from testifycompat import setup
from testifycompat import teardown
from testifycompat import TestCase
from tests.testingutils import autospec_method
from tron import mcp
from tron.config import config_parse
from tron.config import manager
from tron.core.job_collection import JobCollection
from tron.serialize.runstate import statemanager


class TestMasterControlProgram(TestCase):

    TEST_CONFIG = 'tests/data/test_config.yaml'

    @setup
    def setup_mcp(self):
        self.working_dir = tempfile.mkdtemp()
        self.config_path = tempfile.mkdtemp()
        self.mcp = mcp.MasterControlProgram(
            self.working_dir,
            self.config_path,
        )
        self.mcp.state_watcher = mock.create_autospec(
            statemanager.StateChangeWatcher,
        )

    @teardown
    def teardown_mcp(self):
        shutil.rmtree(self.config_path)
        shutil.rmtree(self.working_dir)

    def test_reconfigure(self):
        autospec_method(self.mcp._load_config)
        self.mcp.state_watcher = mock.MagicMock()
        self.mcp.reconfigure()
        self.mcp._load_config.assert_called_with(reconfigure=True)

    def test_load_config(self):
        autospec_method(self.mcp.apply_config)
        self.mcp.config = mock.create_autospec(manager.ConfigManager)
        self.mcp._load_config()
        self.mcp.state_watcher.disabled.assert_called_with()
        self.mcp.apply_config.assert_called_with(
            self.mcp.config.load.return_value,
            reconfigure=False,
        )

    @mock.patch('tron.mcp.MesosClusterRepository', autospec=True)
    @mock.patch('tron.mcp.node.NodePoolRepository', autospec=True)
    def test_apply_config(self, mock_repo, mock_cluster_repo):
        config_container = mock.create_autospec(config_parse.ConfigContainer)
        master_config = config_container.get_master.return_value
        autospec_method(self.mcp.apply_collection_config)
        autospec_method(self.mcp.build_job_scheduler_factory)
        self.mcp.apply_config(config_container)
        self.mcp.state_watcher.update_from_config.assert_called_with(
            master_config.state_persistence,
        )
        assert_equal(self.mcp.context.base, master_config.command_context)
        assert_equal(len(self.mcp.apply_collection_config.mock_calls), 1)
        mock_repo.update_from_config.assert_called_with(
            master_config.nodes,
            master_config.node_pools,
            master_config.ssh_options,
        )
        mock_cluster_repo.configure.assert_called_with(
            master_config.mesos_options,
        )
        self.mcp.build_job_scheduler_factory(master_config)

    def test_update_state_watcher_config_changed(self):
        self.mcp.state_watcher.update_from_config.return_value = True
        self.mcp.jobs = mock.create_autospec(JobCollection)
        self.mcp.jobs.__iter__.return_values = {
            'a': mock.Mock(),
            'b': mock.Mock(),
        }
        state_config = mock.Mock()
        self.mcp.update_state_watcher_config(state_config)
        self.mcp.state_watcher.update_from_config.assert_called_with(
            state_config,
        )
        assert_equal(
            self.mcp.state_watcher.save_job.mock_calls,
            [mock.call(j.job) for j in self.mcp.jobs],
        )

    def test_update_state_watcher_config_no_change(self):
        self.mcp.state_watcher.update_from_config.return_value = False
        self.mcp.jobs = {'a': mock.Mock(), 'b': mock.Mock()}
        state_config = mock.Mock()
        self.mcp.update_state_watcher_config(state_config)
        assert not self.mcp.state_watcher.save_job.mock_calls


class TestMasterControlProgramRestoreState(TestCase):
    @setup
    def setup_mcp(self):
        self.working_dir = tempfile.mkdtemp()
        self.config_path = tempfile.mkdtemp()
        self.mcp = mcp.MasterControlProgram(
            self.working_dir,
            self.config_path,
        )
        self.mcp.jobs = mock.create_autospec(JobCollection)
        self.mcp.state_watcher = mock.create_autospec(
            statemanager.StateChangeWatcher,
        )

    @teardown
    def teardown_mcp(self):
        shutil.rmtree(self.working_dir)
        shutil.rmtree(self.config_path)

    @mock.patch('tron.mcp.MesosClusterRepository', autospec=True)
    def test_restore_state(self, mock_cluster_repo):
        job_state_data = {'1': 'things', '2': 'things'}
        mesos_state_data = {'3': 'things', '4': 'things'}
        state_data = {
            'mesos_state': mesos_state_data,
            'job_state': job_state_data
        }
        self.mcp.state_watcher.restore.return_value = state_data
        action_runner = mock.Mock()
        self.mcp.restore_state(action_runner)
        mock_cluster_repo.restore_state.assert_called_with(mesos_state_data, )
        self.mcp.jobs.restore_state.assert_called_with(
            job_state_data, action_runner
        )


if __name__ == '__main__':
    run()



from __future__ import absolute_import
from __future__ import unicode_literals

import datetime

import mock

from testifycompat import assert_equal
from testifycompat import assert_raises
from testifycompat import run
from testifycompat import setup
from testifycompat import TestCase
from tron import command_context
from tron import node
from tron import scheduler
from tron.core import actionrun
from tron.core import job
from tron.core import jobrun
from tron.core.jobrun import JobRunCollection


class TestEmptyContext(TestCase):
    @setup
    def build_context(self):
        self.context = command_context.CommandContext(None)

    def test__getitem__(self):
        assert_raises(KeyError, self.context.__getitem__, 'foo')

    def test_get(self):
        assert not self.context.get('foo')


class TestBuildFilledContext(TestCase):
    def test_build_filled_context_no_objects(self):
        output = command_context.build_filled_context()
        assert not output.base
        assert not output.next

    def test_build_filled_context_single(self):
        output = command_context.build_filled_context(
            command_context.JobContext,
        )
        assert isinstance(output.base, command_context.JobContext)
        assert not output.next

    def test_build_filled_context_chain(self):
        objs = [command_context.JobContext, command_context.JobRunContext]
        output = command_context.build_filled_context(*objs)
        assert isinstance(output.base, objs[1])
        assert isinstance(output.next.base, objs[0])
        assert not output.next.next


class SimpleContextTestCaseBase(TestCase):
    __test__ = False

    def test_hit(self):
        assert_equal(self.context['foo'], 'bar')

    def test_miss(self):
        assert_raises(KeyError, self.context.__getitem__, 'your_mom')

    def test_get_hit(self):
        assert_equal(self.context.get('foo'), 'bar')

    def test_get_miss(self):
        assert not self.context.get('unknown')


class SimpleDictContextTestCase(SimpleContextTestCaseBase):
    @setup
    def build_context(self):
        self.context = command_context.CommandContext(dict(foo='bar'))


class SimpleObjectContextTestCase(SimpleContextTestCaseBase):
    @setup
    def build_context(self):
        class Obj(object):
            foo = 'bar'

        self.context = command_context.CommandContext(Obj)


class ChainedDictContextTestCase(SimpleContextTestCaseBase):
    @setup
    def build_context(self):
        self.next_context = command_context.CommandContext(
            dict(foo='bar', next_foo='next_bar'),
        )
        self.context = command_context.CommandContext(
            dict(),
            self.next_context,
        )

    def test_chain_get(self):
        assert_equal(self.context['next_foo'], 'next_bar')


class ChainedDictOverrideContextTestCase(SimpleContextTestCaseBase):
    @setup
    def build_context(self):
        self.next_context = command_context.CommandContext(
            dict(foo='your mom', next_foo='next_bar'),
        )
        self.context = command_context.CommandContext(
            dict(foo='bar'),
            self.next_context,
        )

    def test_chain_get(self):
        assert_equal(self.context['next_foo'], 'next_bar')


class ChainedObjectOverrideContextTestCase(SimpleContextTestCaseBase):
    @setup
    def build_context(self):
        class MyObject(TestCase):
            pass

        obj = MyObject()
        obj.foo = 'bar'

        self.next_context = command_context.CommandContext(
            dict(foo='your mom', next_foo='next_bar'),
        )
        self.context = command_context.CommandContext(obj, self.next_context)

    def test_chain_get(self):
        assert_equal(self.context['next_foo'], 'next_bar')


class TestJobContext(TestCase):
    @setup
    def setup_job(self):
        self.last_success = mock.Mock(run_time=datetime.datetime(2012, 3, 14))
        mock_scheduler = mock.create_autospec(scheduler.ConstantScheduler)
        run_collection = mock.create_autospec(
            JobRunCollection,
            last_success=self.last_success,
        )
        self.job = job.Job(
            "jobname",
            mock_scheduler,
            run_collection=run_collection,
        )
        self.context = command_context.JobContext(self.job)

    def test_name(self):
        assert_equal(self.context.name, self.job.name)

    def test__getitem__last_success(self):
        item = self.context["last_success#day-1"]
        expected_date = self.last_success.run_time - datetime.timedelta(days=1)
        assert_equal(item, str(expected_date.day))

        item = self.context["last_success#shortdate"]
        assert_equal(item, "2012-03-14")

    def test__getitem__last_success_bad_date_spec(self):
        name = "last_success#beers-3"
        assert_raises(KeyError, lambda: self.context[name])

    def test__getitem__last_success_bad_date_name(self):
        name = "first_success#shortdate-1"
        assert_raises(KeyError, lambda: self.context[name])

    def test__getitem__last_success_no_date_spec(self):
        name = "last_success"
        assert_raises(KeyError, lambda: self.context[name])

    def test__getitem__missing(self):
        assert_raises(KeyError, lambda: self.context['bogus'])


class TestJobRunContext(TestCase):
    @setup
    def setup_context(self):
        self.jobrun = mock.create_autospec(jobrun.JobRun, run_time='sometime')
        self.context = command_context.JobRunContext(self.jobrun)

    def test_cleanup_job_status(self):
        self.jobrun.action_runs.is_failed = False
        self.jobrun.action_runs.is_complete_without_cleanup = True
        assert_equal(self.context.cleanup_job_status, 'SUCCESS')

    def test_cleanup_job_status_failure(self):
        self.jobrun.action_runs.is_failed = True
        assert_equal(self.context.cleanup_job_status, 'FAILURE')

    def test_runid(self):
        assert_equal(self.context.runid, self.jobrun.id)

    @mock.patch('tron.command_context.timeutils.DateArithmetic', autospec=True)
    def test__getitem__(self, mock_date_math):
        name = 'date_name'
        time_value = self.context[name]
        mock_date_math.parse.assert_called_with(name, self.jobrun.run_time)
        assert_equal(time_value, mock_date_math.parse.return_value)


class TestActionRunContext(TestCase):
    @setup
    def build_context(self):
        mock_node = mock.create_autospec(node.Node, hostname='something')
        self.action_run = mock.create_autospec(
            actionrun.ActionRun,
            action_name='something',
            node=mock_node,
        )
        self.context = command_context.ActionRunContext(self.action_run)

    def test_actionname(self):
        assert_equal(self.context.actionname, self.action_run.action_name)

    def test_node_hostname(self):
        assert_equal(self.context.node, self.action_run.node.hostname)


class TestFiller(TestCase):
    @setup
    def setup_filler(self):
        self.filler = command_context.Filler()

    def test_filler_with_job__getitem__(self):
        context = command_context.JobContext(self.filler)
        todays_date = datetime.date.today().strftime("%Y-%m-%d")
        assert_equal(context['last_success#shortdate'], todays_date)

    def test_filler_with_job_run__getitem__(self):
        context = command_context.JobRunContext(self.filler)
        todays_date = datetime.date.today().strftime("%Y-%m-%d")
        assert_equal(context['shortdate'], todays_date)


if __name__ == '__main__':
    run()



from __future__ import absolute_import
from __future__ import unicode_literals

import mock
from twisted.python import failure

from testifycompat import assert_equal
from testifycompat import assert_not_equal
from testifycompat import setup
from testifycompat import TestCase
from tests.testingutils import autospec_method
from tron import ssh


class TestClientTransport(TestCase):
    @setup
    def setup_transport(self):
        self.username = 'username'
        self.options = mock.Mock()
        self.expected_pub_key = mock.Mock()
        self.transport = ssh.ClientTransport(
            self.username,
            self.options,
            self.expected_pub_key,
        )

    def test_verifyHostKey_missing_pub_key(self):
        self.transport.expected_pub_key = None
        result = self.transport.verifyHostKey(mock.Mock(), mock.Mock())
        assert_equal(result.result, 1)

    @mock.patch('tron.ssh.keys', autospec=True)
    def test_verifyHostKey_matching_pub_key(self, mock_keys):
        mock_keys.Key.fromString.return_value = self.expected_pub_key
        public_key = mock.Mock()
        result = self.transport.verifyHostKey(public_key, mock.Mock())
        assert_equal(result.result, 2)
        mock_keys.Key.fromString.assert_called_with(public_key)

    @mock.patch('tron.ssh.keys', autospec=True)
    def test_verifyHostKey_mismatch_pub_key(self, _):
        public_key = mock.Mock()
        result = self.transport.verifyHostKey(public_key, mock.Mock())
        assert isinstance(result.result, failure.Failure)

    def test_connnectionSecure(self):
        self.transport.connection_defer = mock.Mock()
        autospec_method(self.transport.requestService)
        self.transport.connectionSecure()
        conn = self.transport.connection_defer.mock_calls[0][1][0]
        assert isinstance(conn, ssh.ClientConnection)
        auth_service = self.transport.requestService.mock_calls[0][1][0]
        assert isinstance(auth_service, ssh.NoPasswordAuthClient)


class TestSSHAuthOptions(TestCase):
    def test_from_config_none(self):
        ssh_conf = mock.Mock(agent=False, identities=[])
        ssh_options = ssh.SSHAuthOptions.from_config(ssh_conf)
        assert_equal(ssh_options['noagent'], True)
        assert_equal(ssh_options.identitys, [])

    def test_from_config_both(self):
        identities = ['one', 'two']
        ssh_conf = mock.Mock(agent=True, identities=identities)
        ssh_options = ssh.SSHAuthOptions.from_config(ssh_conf)
        assert_equal(ssh_options['noagent'], False)
        assert_equal(ssh_options.identitys, identities)

    def test__eq__true(self):
        config = mock.Mock(agent=True, identities=['one', 'two'])
        assert_equal(
            ssh.SSHAuthOptions.from_config(config),
            ssh.SSHAuthOptions.from_config(config),
        )

    def test__eq__false(self):
        config = mock.Mock(agent=True, identities=['one', 'two'])
        second_config = mock.Mock(agent=True, identities=['two'])
        assert_not_equal(
            ssh.SSHAuthOptions.from_config(config),
            ssh.SSHAuthOptions.from_config(second_config),
        )
